"""Mutual TLS security scheme."""

from __future__ import annotations

from typing import Any, Sequence

from tigrbl_base._base._security_base import OpenAPISecurityDependency


class MutualTLS(OpenAPISecurityDependency):
    def __init__(
        self,
        *,
        scheme_name: str = "MutualTLS",
        description: str | None = None,
        scopes: Sequence[str] | None = None,
        auto_error: bool = True,
    ) -> None:
        payload: dict[str, Any] = {"type": "mutualTLS"}
        if description is not None:
            payload["description"] = description
        super().__init__(
            scheme_name=scheme_name,
            scheme=payload,
            scopes=scopes,
            auto_error=auto_error,
        )



# tigrbl/v3/schema/decorators.py
from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional

from pydantic import BaseModel, create_model

from tigrbl_core.config.constants import TIGRBL_SCHEMA_DECLS_ATTR

logger = logging.getLogger("uvicorn")


@dataclass(frozen=True)
class _SchemaDecl:
    alias: str  # name under model.schemas.<alias>
    kind: str  # "in" | "out"


def _coerce_schema_model(schema_cls: type, alias: str, kind: str) -> type:
    """Promote plain classes to Pydantic models for schema declarations."""
    del alias, kind
    if isinstance(schema_cls, type) and issubclass(schema_cls, BaseModel):
        return schema_cls

    annotations = dict(getattr(schema_cls, "__annotations__", {}) or {})
    fields: dict[str, tuple[Any, Any]] = {}
    for field_name, field_type in annotations.items():
        default = getattr(schema_cls, field_name, ...)
        fields[field_name] = (field_type, default)

    model_name = f"{schema_cls.__name__}Schema"
    return create_model(
        model_name, __base__=BaseModel, __module__=schema_cls.__module__, **fields
    )


def _register_schema_decl(
    target_model: type, alias: str, kind: str, schema_cls: type
) -> None:
    """Store schema declarations on the model for later binding."""
    logger.debug(
        "Registering schema decl for %s alias=%s kind=%s",
        target_model.__name__,
        alias,
        kind,
    )
    if kind not in ("in", "out"):
        logger.debug("Invalid kind '%s' for schema decl", kind)
        raise ValueError("schema_ctx(kind=...) must be 'in' or 'out'")
    mapping: Dict[str, Dict[str, type]] = (
        getattr(target_model, TIGRBL_SCHEMA_DECLS_ATTR, None) or {}
    )
    bucket = dict(mapping.get(alias, {}))
    bucket[kind] = schema_cls
    mapping[alias] = bucket
    setattr(target_model, TIGRBL_SCHEMA_DECLS_ATTR, mapping)
    logger.debug("Registered schema %s for alias '%s'", schema_cls, alias)


def schema_ctx(*, alias: str, kind: str = "out", for_: Optional[type] = None):
    """Register a named schema for a model."""

    def deco(schema_cls: type):
        if not isinstance(schema_cls, type):
            logger.debug("schema_ctx applied to non-class %r", schema_cls)
            raise TypeError("@schema_ctx must decorate a class")

        normalized_schema = _coerce_schema_model(schema_cls, alias, kind)
        setattr(
            normalized_schema,
            "__tigrbl_schema_decl__",
            _SchemaDecl(alias=alias, kind=kind),
        )

        if for_ is not None:
            logger.debug(
                "Registering schema %s for model %s via decorator",
                normalized_schema,
                for_.__name__,
            )
            _register_schema_decl(for_, alias, kind, normalized_schema)

        logger.debug(
            "Attached schema decl to %s: alias=%s kind=%s",
            normalized_schema,
            alias,
            kind,
        )
        return normalized_schema

    return deco


__all__ = ["schema_ctx"]



"""HTTP bearer security scheme."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Sequence

from tigrbl_runtime.runtime.status.exceptions import HTTPException
from tigrbl_runtime.runtime.status.mappings import status
from tigrbl_base._base._security_base import OpenAPISecurityDependency


@dataclass(frozen=True)
class HTTPAuthorizationCredentials:
    scheme: str
    credentials: str


class HTTPBearer(OpenAPISecurityDependency):
    def __init__(
        self,
        auto_error: bool = True,
        *,
        scheme_name: str = "HTTPBearer",
        scheme: str = "bearer",
        bearer_format: str | None = None,
        description: str | None = None,
        scopes: Sequence[str] | None = None,
    ) -> None:
        payload: dict[str, Any] = {"type": "http", "scheme": scheme}
        if bearer_format is not None:
            payload["bearerFormat"] = bearer_format
        if description is not None:
            payload["description"] = description
        super().__init__(
            scheme_name=scheme_name,
            scheme=payload,
            scopes=scopes,
            auto_error=auto_error,
        )
        self.http_scheme = scheme

    def __call__(self, request: Any) -> HTTPAuthorizationCredentials | None:
        header = request.headers.get("authorization")
        if not header:
            if self.auto_error:
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail="Forbidden",
                )
            return None

        try:
            incoming_scheme, credentials = header.split(" ", 1)
        except ValueError:
            if self.auto_error:
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail="Forbidden",
                )
            return None

        if incoming_scheme.lower() != self.http_scheme.lower():
            if self.auto_error:
                raise HTTPException(
                    status_code=status.HTTP_403_FORBIDDEN,
                    detail="Forbidden",
                )
            return None

        return HTTPAuthorizationCredentials(
            scheme=incoming_scheme,
            credentials=credentials,
        )



from tigrbl_runtime.runtime.kernel import _default_kernel, build_phase_chains
from .router import mount_diagnostics
from .methodz import build_methodz_endpoint as _build_methodz_endpoint
from .hookz import build_hookz_endpoint as _build_hookz_endpoint
from .kernelz import build_kernelz_endpoint as _build_kernelz_endpoint
from .utils import (
    table_iter as _table_iter,
    model_iter as _model_iter,
    opspecs as _opspecs,
    label_callable as _label_callable,
    label_hook as _label_hook,
)

__all__ = [
    "mount_diagnostics",
    "_build_methodz_endpoint",
    "_build_hookz_endpoint",
    "_build_kernelz_endpoint",
    "_table_iter",
    "_model_iter",
    "_opspecs",
    "_label_callable",
    "_label_hook",
    "build_phase_chains",
    "_default_kernel",
]



"""Concrete file response primitive."""

from __future__ import annotations

import mimetypes
from pathlib import Path

from ._response import Response


class FileResponse(Response):
    def __init__(self, path: str, media_type: str | None = None) -> None:
        payload = Path(path).read_bytes()
        content_type = (
            media_type or mimetypes.guess_type(path)[0] or "application/octet-stream"
        )
        super().__init__(
            status_code=200,
            headers=[("content-type", content_type)],
            body=payload,
            media_type=content_type,
        )
        self.path = str(path)


__all__ = ["FileResponse"]



# tigrbl/_concrete/_schema.py
from __future__ import annotations

from dataclasses import dataclass
from typing import Type

try:  # pragma: no cover
    from pydantic import BaseModel  # type: ignore
except Exception:  # pragma: no cover

    class BaseModel:  # minimal stub for typing only
        pass


from tigrbl_base._base import SchemaBase
from tigrbl_core._spec.schema_spec import SchemaKind


@dataclass(frozen=True, slots=True)
class Schema(SchemaBase):
    """Concrete schema paired with its alias and kind."""

    model: Type[BaseModel]
    kind: SchemaKind = "out"
    alias: str | None = None


__all__ = ["Schema"]



"""Hook-related decorators for Tigrbl v3."""

from __future__ import annotations

from enum import Enum
from typing import Iterable, Union

from tigrbl_core.config.constants import HOOK_DECLS_ATTR
from tigrbl_concrete._concrete import Hook
from tigrbl_runtime.runtime.exceptions import InvalidHookPhaseError
from tigrbl_atoms import HookPhase, HookPhases


def hook_ctx(ops: Union[str, Iterable[str]], *, phase: str | HookPhase | Enum):
    """Declare a ctx-only hook for one/many ops at a given phase."""

    phase_value = phase.value if isinstance(phase, Enum) else phase
    try:
        normalized_phase = HookPhase(phase_value)
    except ValueError as exc:
        raise InvalidHookPhaseError(
            phase=str(phase_value),
            allowed_phases=tuple(p.value for p in HookPhases),
        ) from exc

    def deco(fn):
        from .op import _ensure_cm, _unwrap

        cm = _ensure_cm(fn)
        f = _unwrap(cm)
        f.__tigrbl_ctx_only__ = True
        lst = getattr(f, HOOK_DECLS_ATTR, [])
        lst.append(Hook(phase=normalized_phase, fn=f, ops=ops))
        setattr(f, HOOK_DECLS_ATTR, lst)
        return cm

    return deco


__all__ = ["hook_ctx", "HOOK_DECLS_ATTR"]



from __future__ import annotations

import logging
from typing import Any, Awaitable, Callable, Dict, Mapping, Optional

from tigrbl_base._base._rpc_map import (
    _coerce_payload,
    _serialize_output,
    _validate_input,
)
from ..._concrete import engine_resolver as _resolver
from tigrbl_runtime.runtime import executor as _executor

logger = logging.getLogger("uvicorn")
logger.debug("Loaded module v3/mapping/router/resource_proxy")


class _ResourceProxy:
    """Dynamic proxy that resolves operation payloads for core operations."""

    __slots__ = ("_model", "_serialize", "_router")

    def __init__(
        self, model: type, *, serialize: bool = True, router: Any = None
    ) -> None:  # pragma: no cover - trivial
        self._model = model
        self._serialize = serialize
        self._router = router

    def __repr__(self) -> str:  # pragma: no cover - debug helper
        return f"<ResourceProxy {self._model.__name__}>"

    def __getattr__(self, alias: str) -> Callable[..., Awaitable[Any]]:
        logger.debug("Resolving core handler '%s' for %s", alias, self._model.__name__)
        handlers_root = getattr(self._model, "handlers", None)
        h_alias = getattr(handlers_root, alias, None) if handlers_root else None
        has_core_handler = h_alias is not None and hasattr(h_alias, "core")
        ops_root = getattr(self._model, "ops", None)
        by_alias = getattr(ops_root, "by_alias", None)
        has_op_alias = isinstance(by_alias, dict) and alias in by_alias

        if not has_core_handler and not has_op_alias:
            logger.debug(
                "No core handler '%s' found for %s", alias, self._model.__name__
            )
            raise AttributeError(f"{self._model.__name__} has no core method '{alias}'")

        async def _call(
            payload: Any = None,
            *,
            db: Any | None = None,
            request: Any = None,
            ctx: Optional[Dict[str, Any]] = None,
        ) -> Any:
            raw_payload = _coerce_payload(payload)
            logger.debug(
                "Preparing %s.%s with payload %s",
                self._model.__name__,
                alias,
                raw_payload,
            )
            if alias == "bulk_delete" and not isinstance(raw_payload, Mapping):
                raw_payload = {"ids": raw_payload}
                logger.debug("Coerced bulk_delete payload to mapping: %s", raw_payload)
            norm_payload = _validate_input(self._model, alias, alias, raw_payload)

            seed_ctx: Dict[str, Any] = dict(ctx or {})

            release_db = None
            if db is None:
                db, release_db = _resolver.acquire(
                    router=self._router, model=self._model, op_alias=alias
                )

            seed_ctx.setdefault("payload", norm_payload)
            seed_ctx.setdefault("db", db)
            if request is not None:
                seed_ctx.setdefault("request", request)
            app_ref = (
                getattr(request, "app", None)
                or seed_ctx.get("app")
                or self._router
                or self._model
            )
            seed_ctx.setdefault("app", app_ref)
            seed_ctx.setdefault("router", seed_ctx.get("router") or app_ref)
            seed_ctx.setdefault("model", self._model)
            seed_ctx.setdefault("op", alias)
            seed_ctx.setdefault("method", alias)
            seed_ctx.setdefault("target", alias)

            try:
                if self._serialize:
                    seed_ctx["response_serializer"] = lambda result: _serialize_output(
                        self._model, alias, alias, result
                    )
                invoke_op = getattr(_executor, "invoke_op", None)
                if not callable(invoke_op):
                    from tigrbl_runtime.runtime.executor.invoke import (
                        invoke_op as _invoke_op,
                    )

                    invoke_op = _invoke_op
                return await invoke_op(
                    request=request,
                    db=db,
                    model=self._model,
                    alias=alias,
                    ctx=seed_ctx,
                )
            finally:
                if release_db is not None:
                    try:
                        release_db()
                    except Exception:
                        logger.debug("Error releasing core proxy DB", exc_info=True)

        _call.__name__ = f"{self._model.__name__}.{alias}"
        _call.__qualname__ = _call.__name__
        _call.__doc__ = f"Helper for core call {self._model.__name__}.{alias}"
        return _call



"""Concrete runtime hook wrapper for Tigrbl v3."""

from __future__ import annotations

from tigrbl_base._base._hook_base import HookBase


class Hook(HookBase):
    """Concrete hook implementation."""


__all__ = ["Hook"]



from __future__ import annotations

import asyncio
import inspect
from pathlib import Path
from typing import Any, Dict, Iterable, Mapping, Optional, Sequence

from tigrbl_concrete._concrete import engine_resolver as _resolver

try:  # pragma: no cover
    from sqlalchemy import event, text
    from sqlalchemy.engine import Engine
    from sqlalchemy.schema import CreateSchema  # type: ignore
except Exception:  # pragma: no cover
    event = text = CreateSchema = None  # type: ignore
    Engine = object  # type: ignore

from tigrbl_core.config.constants import __SAFE_IDENT__


class _NoopAwaitable:
    def __await__(self):
        if False:
            yield None
        return None


__all__ = [
    "register_sqlite_attach",
    "ensure_schemas",
    "bootstrap_dbschema",
    "sqlite_default_attach_map",
    "initialize",
]


# ---------------------------------------------------------------------------
# SQLite helpers
# ---------------------------------------------------------------------------


def _quote_ident_sqlite(name: str) -> str:
    if __SAFE_IDENT__.match(name or ""):
        return name
    return '"' + (name or "").replace('"', '""') + '"'


def _attached_names_sqlite(dbapi_conn: Any) -> set[str]:
    names: set[str] = set()
    cur = None
    try:
        cur = dbapi_conn.cursor()
        cur.execute("PRAGMA database_list")
        for row in cur.fetchall():
            names.add(str(row[1]))
    finally:
        try:
            cur and cur.close()
        except Exception:
            pass
    return names


def _attach_sqlite_dbapi(dbapi_conn: Any, attachments: Mapping[str, str]) -> None:
    cur = None
    try:
        existing = _attached_names_sqlite(dbapi_conn)
        cur = dbapi_conn.cursor()
        try:
            cur.execute("PRAGMA foreign_keys=ON")
        except Exception:
            pass
        for schema, path in (attachments or {}).items():
            if not path or schema in existing:
                continue
            ident = _quote_ident_sqlite(schema)
            try:
                cur.execute(f"ATTACH DATABASE ? AS {ident}", (path,))
            except Exception:
                cur.execute(f"ATTACH DATABASE '{path}' AS {ident}")  # nosec: application-provided paths
    finally:
        try:
            cur and cur.close()
        except Exception:
            pass


# ---------------------------------------------------------------------------
# Public helpers
# ---------------------------------------------------------------------------


def sqlite_default_attach_map(engine: Engine, schemas: Iterable[str]) -> Dict[str, str]:
    """Return a deterministic SQLite ATTACH map for ``schemas``."""
    sync_engine = getattr(engine, "sync_engine", engine)
    db = getattr(getattr(sync_engine, "url", None), "database", None) or ":memory:"
    if db == ":memory:" or str(db).startswith("file::memory:"):
        return {s: ":memory:" for s in schemas}
    p = Path(db)
    suffix = p.suffix if p.suffix else ".db"
    return {s: str(p.with_name(f"{p.stem}__{s}{suffix}")) for s in schemas}


def register_sqlite_attach(engine: Engine, attachments: Mapping[str, str]) -> Any:
    sync_engine = getattr(engine, "sync_engine", engine)
    if (
        not hasattr(sync_engine, "dialect")
        or getattr(sync_engine.dialect, "name", "") != "sqlite"
    ):
        return None

    def _connect_listener(dbapi_conn, _):  # type: ignore[override]
        try:
            _attach_sqlite_dbapi(dbapi_conn, attachments)
        except Exception:
            pass

    event.listen(sync_engine, "connect", _connect_listener)
    return _connect_listener


def ensure_schemas(engine: Engine, schemas: Iterable[str]) -> Sequence[str]:
    if not schemas:
        return tuple()
    if not hasattr(engine, "dialect"):
        return tuple(schemas)

    dialect = getattr(engine.dialect, "name", "")
    attempted: list[str] = []

    if dialect == "sqlite":
        return tuple()

    if text is None:  # pragma: no cover
        raise RuntimeError("SQLAlchemy is required for ensure_schemas")

    try:
        with engine.begin() as conn:
            for name in dict.fromkeys(schemas).keys():
                if not name:
                    continue
                attempted.append(name)
                try:
                    conn.execute(CreateSchema(name))  # type: ignore[arg-type]
                except Exception:
                    try:
                        if dialect in ("postgresql", "redshift"):
                            conn.execute(text(f'CREATE SCHEMA IF NOT EXISTS "{name}"'))
                        elif dialect in ("mysql", "mariadb"):
                            conn.execute(text(f"CREATE SCHEMA IF NOT EXISTS `{name}`"))
                        elif dialect in ("mssql", "sqlserver"):
                            conn.execute(
                                text(
                                    "IF NOT EXISTS (SELECT * FROM sys.schemas WHERE name = :n) "
                                    "EXEC('CREATE SCHEMA ' + QUOTENAME(:n))"
                                ),
                                {"n": name},
                            )
                        else:
                            conn.execute(text(f"CREATE SCHEMA IF NOT EXISTS {name}"))
                    except Exception:
                        pass
    except Exception:
        pass

    return tuple(attempted)


def bootstrap_dbschema(
    engine: Engine,
    *,
    schemas: Optional[Iterable[str]] = None,
    sqlite_attachments: Optional[Mapping[str, str]] = None,
    immediate: bool = True,
) -> Dict[str, Any]:
    attempted = tuple()
    if schemas:
        attempted = ensure_schemas(engine, schemas)

    listener = None
    if sqlite_attachments:
        listener = register_sqlite_attach(engine, sqlite_attachments)
        sync_engine = getattr(engine, "sync_engine", engine)
        if immediate and getattr(sync_engine.dialect, "name", "") == "sqlite":
            try:
                with sync_engine.connect() as conn:
                    dbapi = getattr(conn, "connection", None)
                    if dbapi is not None:
                        _attach_sqlite_dbapi(dbapi, sqlite_attachments)
            except Exception:
                pass

    return {
        "attempted_schemas": attempted,
        "sqlite_attachments": dict(sqlite_attachments or {}),
        "listener": listener,
    }


# ---------------------------------------------------------------------------
# Internal creation helper
# ---------------------------------------------------------------------------


def _create_all_on_bind(
    bind,
    *,
    schemas: Iterable[str] | None = None,
    sqlite_attachments: Mapping[str, str] | None = None,
    tables: Iterable[Any] | None = None,
) -> None:
    engine = getattr(bind, "engine", bind)
    tables = list(tables or [])

    if not hasattr(engine, "dialect"):
        return _NoopAwaitable()

    schema_names = set(schemas or [])
    for t in tables:
        if getattr(t, "schema", None):
            schema_names.add(t.schema)

    attachments = sqlite_attachments
    if attachments is None and getattr(engine.dialect, "name", "") == "sqlite":
        if schema_names:
            attachments = sqlite_default_attach_map(engine, schema_names)

    if attachments:
        bootstrap_dbschema(
            engine,
            schemas=schema_names,
            sqlite_attachments=attachments,
            immediate=True,
        )
    else:
        ensure_schemas(engine, schema_names)

    by_meta: dict[Any, list[Any]] = {}
    for t in tables:
        by_meta.setdefault(t.metadata, []).append(t)
    for md, group in by_meta.items():
        md.create_all(bind=bind, checkfirst=True, tables=group)


# ---------------------------------------------------------------------------
# Public initialize
# ---------------------------------------------------------------------------


def initialize(
    obj: Any,
    *,
    schemas: Iterable[str] | None = None,
    sqlite_attachments: Mapping[str, str] | None = None,
    tables: Iterable[Any] | None = None,
):
    if getattr(obj, "_ddl_executed", False):
        return

    ts = list(tables or [])
    if not ts:
        if hasattr(obj, "_collect_tables"):
            ts = list(obj._collect_tables())  # type: ignore[attr-defined]
        elif hasattr(obj, "__table__"):
            ts = [obj.__table__]  # type: ignore[attr-defined]

    provider_tables: list[tuple[Any, list[Any]]] = []
    model_registry = getattr(obj, "tables", None)
    model_values = (
        tuple(model_registry.values())
        if isinstance(model_registry, Mapping)
        else tuple(model_registry or ())
    )
    if model_values:
        grouped: dict[int, tuple[Any, list[Any]]] = {}
        for model in model_values:
            table = getattr(model, "__table__", None)
            if table is None:
                continue
            prov_for_model = _resolver.resolve_provider(router=obj, model=model)
            if prov_for_model is None:
                continue
            key = id(prov_for_model)
            entry = grouped.get(key)
            if entry is None:
                grouped[key] = (prov_for_model, [table])
            else:
                entry[1].append(table)
        provider_tables = list(grouped.values())

    kwargs: Dict[str, Any] = {}
    if hasattr(obj, "_collect_tables"):
        kwargs["router"] = obj
    elif hasattr(obj, "__table__"):
        kwargs["model"] = obj

    prov = _resolver.resolve_provider(**kwargs)
    if prov is None and not provider_tables:
        raise ValueError("Engine provider is not configured")

    if not provider_tables and prov is not None:
        provider_tables = [(prov, ts)]

    def _close_without_loop(db):
        close = getattr(db, "close", None)
        if not callable(close):
            return
        out = close()
        if inspect.isawaitable(out):
            asyncio.run(out)

    def _close_with_loop(db):
        close = getattr(db, "close", None)
        if not callable(close):
            return _NoopAwaitable()
        out = close()
        if inspect.isawaitable(out):
            loop = asyncio.get_running_loop()
            return loop.create_task(out)
        return None

    try:
        asyncio.get_running_loop()
    except RuntimeError:
        # No running event loop; fall back to fully synchronous bootstrap
        for active_provider, active_tables in provider_tables:
            db = next(active_provider.get_db())
            try:
                bind = db.get_bind() if hasattr(db, "get_bind") else db
                _create_all_on_bind(
                    bind,
                    schemas=schemas,
                    sqlite_attachments=sqlite_attachments,
                    tables=active_tables,
                )
            finally:
                _close_without_loop(db)
        setattr(obj, "_ddl_executed", True)
        return _NoopAwaitable()
    else:
        # If we're already inside an event loop but the provider is synchronous
        # (i.e. ``get_db`` is neither coroutine nor async generator), we can
        # bootstrap synchronously as well. This mirrors previous "initialize_sync"
        # behaviour and allows ``initialize()`` to be invoked without ``await``
        # from async contexts when using sync engines.
        if not inspect.iscoroutinefunction(
            prov.get_db
        ) and not inspect.isasyncgenfunction(prov.get_db):
            pending_closes = []
            for active_provider, active_tables in provider_tables:
                db = next(active_provider.get_db())
                try:
                    bind = db.get_bind() if hasattr(db, "get_bind") else db
                    _create_all_on_bind(
                        bind,
                        schemas=schemas,
                        sqlite_attachments=sqlite_attachments,
                        tables=active_tables,
                    )
                finally:
                    pending = _close_with_loop(db)
                    if pending is not None:
                        pending_closes.append(pending)
            setattr(obj, "_ddl_executed", True)
            if pending_closes:
                return asyncio.gather(*pending_closes)
            return _NoopAwaitable()

        async def _inner():
            for active_provider, active_tables in provider_tables:
                if inspect.isasyncgenfunction(active_provider.get_db):
                    async for adb in active_provider.get_db():
                        await adb.run_sync(
                            lambda sync_session: _create_all_on_bind(
                                sync_session.get_bind(),
                                schemas=schemas,
                                sqlite_attachments=sqlite_attachments,
                                tables=active_tables,
                            )
                        )
                        break
                else:
                    gen = active_provider.get_db()
                    db = next(gen)
                    try:
                        if hasattr(db, "run_sync"):
                            await db.run_sync(
                                lambda sync_session: _create_all_on_bind(
                                    sync_session.get_bind(),
                                    schemas=schemas,
                                    sqlite_attachments=sqlite_attachments,
                                    tables=active_tables,
                                )
                            )
                        else:
                            bind = db.get_bind()
                            await asyncio.to_thread(
                                _create_all_on_bind,
                                bind,
                                schemas=schemas,
                                sqlite_attachments=sqlite_attachments,
                                tables=active_tables,
                            )
                    finally:
                        try:
                            next(gen)
                        except StopIteration:
                            pass
            setattr(obj, "_ddl_executed", True)

        return _inner()



from __future__ import annotations

import logging
from types import SimpleNamespace
from typing import Any

from tigrbl_base._base import AttrDict

logger = logging.getLogger("uvicorn")
logger.debug("Loaded module v3/mapping/router/common")


# Public type for the Router facade object users pass to include_table(...)
RouterLike = Any


def _default_prefix(model: type) -> str:
    """Default mount prefix for a model router.

    Historically routers were mounted under ``/{resource}``, resulting in
    duplicated path segments such as ``/item/item``.  To expose REST endpoints
    under ``/item`` we now mount routers at the application root by default.
    """
    logger.debug("Default prefix for %s is root '/'", model.__name__)
    return ""


def _has_include_router(obj: Any) -> bool:
    has_router = hasattr(obj, "include_router") and callable(
        getattr(obj, "include_router")
    )
    logger.debug("Object %s has include_router: %s", obj, has_router)
    return has_router


def _mount_router(app_or_router: Any, router: Any, *, prefix: str) -> None:
    """
    Best-effort mount onto a ASGI app or Router.
    If not available, we still attach router under router.routers for later use.
    """
    if app_or_router is None:
        logger.debug("No app/router; skipping mount for prefix %s", prefix)
        return
    try:
        if _has_include_router(app_or_router):
            logger.debug("Mounting router %s at prefix %s", router, prefix)
            app_or_router.include_router(router, prefix=prefix)  # ASGI / Router
        else:
            logger.debug(
                "Provided object %s lacks include_router; not mounting router",
                app_or_router,
            )
    except Exception:
        logger.exception("Failed to mount router at %s", prefix)


def _ensure_router_ns(router: RouterLike) -> None:
    """
    Ensure containers exist on the router facade object.
    """
    for attr, default in (
        ("tables", AttrDict()),
        ("schemas", SimpleNamespace()),
        ("handlers", SimpleNamespace()),
        ("hooks", SimpleNamespace()),
        ("rpc", SimpleNamespace()),
        ("rest", SimpleNamespace()),
        ("routers", {}),
        ("columns", {}),
        ("table_config", {}),
        ("core", SimpleNamespace()),  # helper method proxies
        ("core_raw", SimpleNamespace()),
    ):
        if not hasattr(router, attr):
            setattr(router, attr, default)
            logger.debug("Initialized router.%s", attr)
        else:
            logger.debug("router already has attribute %s", attr)



"""OpenAPI document builder for std routers."""

from __future__ import annotations

from typing import Any

from tigrbl_runtime.runtime.status.mappings import status
from .helpers import (
    _request_schema_from_handler,
    _resolve_component_schema_ref,
    _schema_from_model,
    _security_from_dependencies,
    _security_schemes_from_dependencies,
)


def openapi(router: Any) -> dict[str, Any]:
    paths: dict[str, Any] = {}
    components: dict[str, Any] = {"schemas": {}}
    components_schemas: dict[str, Any] = components["schemas"]

    for route in getattr(router, "_routes", []):
        if not route.include_in_schema:
            continue

        canonical_path = route.path_template.rstrip("/") or "/"
        if canonical_path == "/openrpc.json" or route.name == "openrpc_json":
            continue

        path_item = paths.setdefault(canonical_path, {})
        for method in sorted(route.methods):
            alias = (
                getattr(route, "tigrbl_alias", None)
                or route.name.rsplit(".", maxsplit=1)[-1]
            )
            status_code = route.status_code or (
                status.HTTP_201_CREATED if alias == "create" else status.HTTP_200_OK
            )
            responses: dict[str, Any] = {}
            if route.responses:
                for code, meta in route.responses.items():
                    entry: dict[str, Any] = {
                        "description": meta.get("description", "") or "",
                    }
                    model = meta.get("model")
                    if model is not None:
                        entry["content"] = {
                            "application/json": {
                                "schema": _schema_from_model(
                                    model,
                                    components_schemas,
                                )
                            }
                        }
                    responses[str(code)] = entry
            if str(status_code) not in responses:
                entry = {"description": "Successful Response"}
                schema = route.response_schema
                if schema is None and route.response_model is not None:
                    schema = _schema_from_model(
                        route.response_model,
                        components_schemas,
                    )
                if schema is not None:
                    entry["content"] = {"application/json": {"schema": schema}}
                responses[str(status_code)] = entry

            op: dict[str, Any] = {
                "operationId": route.operation_id or route.name,
                "responses": responses,
            }
            if route.summary:
                op["summary"] = route.summary
            if route.description:
                op["description"] = route.description
            if route.tags:
                op["tags"] = route.tags
            if route.deprecated:
                op["deprecated"] = True

            params: list[dict[str, Any]] = []
            for param_name in route.param_names:
                schema = (route.path_param_schemas or {}).get(param_name) or {
                    "type": "string"
                }
                params.append(
                    {
                        "name": param_name,
                        "in": "path",
                        "required": True,
                        "schema": schema,
                    }
                )

            for qname, qschema in (route.query_param_schemas or {}).items():
                params.append(
                    {
                        "name": qname,
                        "in": "query",
                        "required": bool(qschema.get("required", False)),
                        "schema": {k: v for k, v in qschema.items() if k != "required"},
                    }
                )

            if params:
                op["parameters"] = params

            request_schema = route.request_schema
            if request_schema is None and route.request_model is not None:
                request_schema = _schema_from_model(
                    route.request_model,
                    components_schemas,
                )

            if request_schema is None:
                request_schema = _request_schema_from_handler(route, components_schemas)

            if isinstance(request_schema, dict) and alias.startswith("bulk_"):
                request_schema = _resolve_component_schema_ref(
                    request_schema,
                    components_schemas,
                )

            if request_schema is not None:
                op["requestBody"] = {
                    "required": True,
                    "content": {"application/json": {"schema": request_schema}},
                }

            model = getattr(route, "tigrbl_model", None)
            alias = getattr(route, "tigrbl_alias", None)
            security_deps: list[Any] = []
            security_deps.extend(list(getattr(router, "dependencies", None) or ()))
            security_deps.extend(list(getattr(route, "dependencies", None) or ()))
            security_deps.extend(
                list(getattr(route, "security_dependencies", None) or ())
            )
            if model is not None and isinstance(alias, str):
                specs = getattr(getattr(model, "ops", None), "by_alias", {})
                sp_list = specs.get(alias) or ()
                if sp_list:
                    security_deps.extend(list(getattr(sp_list[0], "secdeps", ()) or ()))

            for dep in tuple(getattr(route, "dependencies", ()) or ()):
                if dep not in security_deps:
                    security_deps.append(dep)

            sec = _security_from_dependencies(security_deps)
            if sec:
                op["security"] = sec
                components.setdefault("securitySchemes", {}).update(
                    _security_schemes_from_dependencies(security_deps)
                )

            path_item[method.lower()] = op

    seen_paths = getattr(router, "_seen_paths", None)
    if isinstance(seen_paths, set):
        for raw_path in seen_paths:
            if not isinstance(raw_path, str):
                continue
            normalized = raw_path.rstrip("/") or "/"
            if normalized in paths:
                continue
            for templated, item in list(paths.items()):
                matcher = _template_matcher(templated)
                if matcher is None or matcher.match(normalized) is None:
                    continue
                paths[normalized] = item
                break

    doc: dict[str, Any] = {
        "openapi": "3.1.0",
        "info": {"title": router.title, "version": router.version},
        "paths": paths,
        "components": components,
    }
    if router.description:
        doc["info"]["description"] = router.description
    return doc


def _template_matcher(path_template: str):
    import re

    escaped = re.escape(path_template)
    pattern = re.sub(r"\\\{[A-Za-z_][A-Za-z0-9_]*\\\}", r"[^/]+", escaped)
    return re.compile(rf"^{pattern}$")



from __future__ import annotations

import inspect
import warnings
from types import SimpleNamespace
from typing import Any, Iterable

from sqlalchemy import text

from tigrbl_runtime.runtime.labels import (
    label_callable as _kernel_label_callable,
    label_hook as _kernel_label_hook,
)


def table_iter(router: Any) -> Iterable[type]:
    tables = getattr(router, "tables", None)
    if isinstance(tables, dict) and tables:
        return tables.values()

    return ()


def model_iter(router: Any) -> Iterable[type]:
    warnings.warn(
        "model_iter is deprecated; use table_iter instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return table_iter(router)


def opspecs(model: type):
    ops = getattr(getattr(model, "opspecs", SimpleNamespace()), "all", ()) or ()
    if ops:
        return ops
    try:
        from tigrbl_core._spec.op_spec import _mro_collect_decorated_ops

        return tuple(_mro_collect_decorated_ops(model))
    except Exception:
        return ()


async def maybe_execute(db: Any, stmt: str):
    try:
        rv = db.execute(text(stmt))  # type: ignore[attr-defined]
        if inspect.isawaitable(rv):
            return await rv
        return rv
    except Exception:
        rv = db.execute(text("select 1"))  # type: ignore[attr-defined]
        if inspect.isawaitable(rv):
            return await rv
        return rv


def label_callable(fn: Any) -> str:
    return _kernel_label_callable(fn)


def label_hook(fn: Any, phase: str) -> str:
    return _kernel_label_hook(fn, phase)



from __future__ import annotations

from typing import Any, Optional
from tigrbl_core._spec.session_spec import SessionSpec, SessionCfg


def _normalize(cfg: Optional[SessionCfg] = None, **kw: Any) -> SessionSpec:
    if cfg is not None and kw:
        raise ValueError("Pass either a mapping/spec or keyword args, not both")
    return SessionSpec.from_any(cfg or kw) or SessionSpec()


def session_ctx(cfg: Optional[SessionCfg] = None, /, **kw: Any):
    """
    Attach a SessionSpec to an App, API, Model/Table, or op handler.

    Precedence is evaluated by the resolver using:
        op > model > router > app
    (Resolver is part of the runtime/engine layer and is independent of this decorator.)
    """
    spec = _normalize(cfg, **kw)

    def _apply(obj: Any) -> Any:
        setattr(obj, "__tigrbl_session_ctx__", spec)
        return obj

    return _apply


def read_only_session(obj: Any = None, /, *, isolation: Optional[str] = None):
    """
    Convenience decorator for read-only sessions.
    """

    def _wrap(o: Any) -> Any:
        setattr(
            o,
            "__tigrbl_session_ctx__",
            SessionSpec(read_only=True, isolation=isolation),
        )
        return o

    return _wrap(obj) if obj is not None else _wrap



"""Router primitives backing ``tigrbl.Router`` and ``tigrbl.App``."""

from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Any, Callable
from types import SimpleNamespace

from tigrbl_base._base import RouterBase
from tigrbl_concrete._concrete import engine_resolver as _resolver
from tigrbl_core._spec.app_spec import _seqify
from tigrbl_core._spec.engine_spec import EngineCfg
from ._table_registry import TableRegistry

from ._routing import (
    add_route as _add_route_impl,
    include_router as _include_router_impl,
    merge_tags,
    normalize_prefix,
    route,
)
from ._httpx import ensure_httpx_sync_transport

from ._route import Route
from tigrbl_concrete.system.docs.openapi.metadata import (
    is_metadata_route as _is_metadata_route_impl,
)

Handler = Callable[..., Any]


def _rest_get(router: Any, path: str, **kwargs: Any):
    return router.route(path, methods=["GET"], **kwargs)


def _rest_post(router: Any, path: str, **kwargs: Any):
    return router.route(path, methods=["POST"], **kwargs)


def _rest_put(router: Any, path: str, **kwargs: Any):
    return router.route(path, methods=["PUT"], **kwargs)


def _rest_patch(router: Any, path: str, **kwargs: Any):
    return router.route(path, methods=["PATCH"], **kwargs)


def _rest_delete(router: Any, path: str, **kwargs: Any):
    return router.route(path, methods=["DELETE"], **kwargs)


@asynccontextmanager
async def _default_lifespan_context(app: Any):
    yield


class Router(RouterBase):
    """API router with transport, dependency, and model/table registry support."""

    TABLES: tuple[Any, ...] = ()
    REST_PREFIX = "/router"
    RPC_PREFIX = "/rpc"
    SYSTEM_PREFIX = "/system"

    @staticmethod
    def _collect_declared_tables(owner: type) -> tuple[Any, ...]:
        collected: list[Any] = []
        seen: set[int] = set()
        for base in owner.__mro__:
            if "TABLES" not in base.__dict__:
                continue
            for table in tuple(base.__dict__.get("TABLES", ()) or ()):
                marker = id(table)
                if marker in seen:
                    continue
                seen.add(marker)
                collected.append(table)
        return tuple(collected)

    def __init__(
        self,
        *,
        engine: EngineCfg | None = None,
        get_db: Callable[..., Any] | None = None,
        title: str = "API",
        version: str = "0.1.0",
        description: str | None = None,
        openapi_url: str = "/openapi.json",
        docs_url: str = "/docs",
        debug: bool = False,
        swagger_ui_version: str = "5.31.0",
        prefix: str = "",
        tags: list[str] | None = None,
        dependencies: list[Any] | None = None,
        include_docs: bool = False,
    ) -> None:
        self.title = title
        self.version = version
        self.description = description
        self.openapi_url = openapi_url
        self.docs_url = docs_url
        self.debug = debug
        self.swagger_ui_version = swagger_ui_version
        class_prefix = getattr(self, "PREFIX", "")
        self.prefix = normalize_prefix(prefix or class_prefix)
        self.tags = list(_seqify(tags))
        self.dependencies = list(_seqify(dependencies))
        # Allow dependencies to be replaced at runtime, typically for testing
        # and environment-specific wiring.
        self.dependency_overrides: dict[Callable[..., Any], Callable[..., Any]] = {}
        self.dependency_overrides_provider = self
        self.lifespan_context = _default_lifespan_context
        if get_db is not None:
            self.get_db = get_db

        self._routes: list[Route] = []
        self.routes = self._routes

        self.name = getattr(self, "NAME", "router")
        self.prefix = normalize_prefix(prefix or getattr(self, "PREFIX", ""))
        self.engine = engine if engine is not None else getattr(self, "ENGINE", None)
        self.tags = list(
            _seqify(tags if tags is not None else getattr(self, "TAGS", ()))
        )
        self.ops = _seqify(getattr(self, "OPS", ()))
        self.schemas = SimpleNamespace()
        self.hooks = SimpleNamespace()
        self.security_deps = _seqify(getattr(self, "SECURITY_DEPS", ()))
        self.deps = _seqify(getattr(self, "DEPS", ()))
        self.response = getattr(self, "RESPONSE", None)
        self.rest_prefix = getattr(self, "REST_PREFIX", "/router")
        self.rpc_prefix = getattr(self, "RPC_PREFIX", "/rpc")
        self.system_prefix = getattr(self, "SYSTEM_PREFIX", "/system")
        self.tables = TableRegistry(
            tables=Router._collect_declared_tables(self.__class__)
        )
        self._table_registry = self.tables
        self._table_regsitry = self.tables

        default_dependencies = list(self.security_deps) + list(self.deps)
        self.dependencies = list(
            _seqify(dependencies if dependencies is not None else default_dependencies)
        )

        _engine_ctx = engine if engine is not None else getattr(self, "ENGINE", None)
        if _engine_ctx is not None:
            _resolver.register_router(self, _engine_ctx)
            _resolver.resolve_provider(router=self)

    def _include_table_impl(
        self,
        table: type,
        *,
        app: Any | None = None,
        prefix: str | None = None,
        mount_router: bool = True,
    ) -> tuple[type, Any]:
        from tigrbl_concrete._mapping.router.include import include_table

        return include_table(
            self,
            table,
            app=app,
            prefix=prefix,
            mount_router=mount_router,
        )

    def _include_tables_impl(
        self,
        tables: tuple[type, ...] | list[type],
        *,
        app: Any | None = None,
        base_prefix: str | None = None,
        mount_router: bool = True,
    ) -> dict[str, Any]:
        from tigrbl_concrete._mapping.router.include import include_tables

        return include_tables(
            self,
            tables,
            app=app,
            base_prefix=base_prefix,
            mount_router=mount_router,
        )

    def _normalize_prefix(self, prefix: str) -> str:
        return normalize_prefix(prefix)

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: object) -> bool:
        return self is other

    @property
    def router(self) -> "Router":
        return self

    def _bump_runtime_plan_revision(self) -> None:
        current = getattr(self, "_runtime_plan_revision", 0)
        try:
            self._runtime_plan_revision = int(current) + 1
        except Exception:
            self._runtime_plan_revision = 1

    def add_route(
        self,
        path: str,
        endpoint: Any,
        *,
        methods: list[str] | tuple[str, ...],
        **kwargs: Any,
    ) -> None:
        _add_route_impl(self, path, endpoint, methods=methods, **kwargs)
        self._bump_runtime_plan_revision()

    def _merge_tags(self, tags: list[str] | None) -> list[str] | None:
        return merge_tags(self.tags, tags)

    def include_router(self, router: Any, *, prefix: str | None = None) -> Any:
        routed = getattr(router, "router", router)
        _include_router_impl(self, routed, prefix=prefix or "")
        self._bump_runtime_plan_revision()
        return router

    def route(
        self, path: str, *, methods: Any, **kwargs: Any
    ) -> Callable[[Handler], Handler]:
        return route(self, path, methods=methods, **kwargs)

    def get(self, path: str, **kwargs: Any) -> Callable[[Handler], Handler]:
        return _rest_get(self, path, **kwargs)

    def post(self, path: str, **kwargs: Any) -> Callable[[Handler], Handler]:
        return _rest_post(self, path, **kwargs)

    def put(self, path: str, **kwargs: Any) -> Callable[[Handler], Handler]:
        return _rest_put(self, path, **kwargs)

    def patch(self, path: str, **kwargs: Any) -> Callable[[Handler], Handler]:
        return _rest_patch(self, path, **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Callable[[Handler], Handler]:
        return _rest_delete(self, path, **kwargs)

    def __call__(self, *args: Any, **kwargs: Any):
        del args, kwargs
        raise TypeError(
            "Router is no longer a transport entrypoint; mount it on TigrblApp."
        )

    async def _execute_route_dependencies(self, route: Route, req: Any) -> None:
        del route, req
        raise RuntimeError("Route dependencies are executed by the runtime/app layer.")

    def _is_metadata_route(self, route: Route) -> bool:
        return _is_metadata_route_impl(self, route)

    async def _execute_dependency_tokens(
        self, kwargs: dict[str, Any], req: Any
    ) -> dict[str, Any]:
        del kwargs, req
        raise RuntimeError("Dependency tokens are executed by the runtime/app layer.")

    async def _invoke_dependency(self, dep: Callable[..., Any], req: Any) -> Any:
        del dep, req
        raise RuntimeError("Dependencies are invoked by the runtime/app layer.")


ensure_httpx_sync_transport()


__all__ = ["Router"]



"""OpenAPI/docs route metadata helpers."""

from __future__ import annotations

from typing import Any


def is_metadata_route(router: Any, route: Any) -> bool:
    if route.name in {"__openapi__", "__docs__"}:
        return True

    openapi_path = (
        router.openapi_url
        if router.openapi_url.startswith("/")
        else f"/{router.openapi_url}"
    )
    docs_path = (
        router.docs_url if router.docs_url.startswith("/") else f"/{router.docs_url}"
    )
    metadata_paths = {router.prefix + openapi_path, router.prefix + docs_path}
    return route.path_template in metadata_paths



"""Backward-compatible engine resolver exports."""

from tigrbl_concrete._concrete.engine_resolver import *  # noqa: F401,F403
from tigrbl_concrete._concrete import engine_resolver as _engine_resolver

_SECRET_KEYS = _engine_resolver._SECRET_KEYS
_hash_secret = _engine_resolver._hash_secret
_spec_key = _engine_resolver._spec_key



"""OAuth2 security scheme."""

from __future__ import annotations

from typing import Any, Mapping, Sequence

from tigrbl_base._base._security_base import OpenAPISecurityDependency


class OAuth2(OpenAPISecurityDependency):
    def __init__(
        self,
        *,
        scheme_name: str = "OAuth2",
        flows: Mapping[str, Any],
        description: str | None = None,
        scopes: Sequence[str] | None = None,
        auto_error: bool = True,
    ) -> None:
        payload: dict[str, Any] = {"type": "oauth2", "flows": dict(flows)}
        if description is not None:
            payload["description"] = description
        super().__init__(
            scheme_name=scheme_name,
            scheme=payload,
            scopes=scopes,
            auto_error=auto_error,
        )



# tigrbl/_concrete/tigrbl_router.py
from __future__ import annotations

import copy
import asyncio
import inspect
from types import SimpleNamespace
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)

from ._router import Router as _Router
from tigrbl_core._spec.engine_spec import EngineCfg
from tigrbl_concrete.ddl import initialize as _ddl_initialize
from tigrbl_concrete._mapping.router.include import (
    include_table as _include_table,
    include_tables as _include_tables,
)
from tigrbl_concrete._mapping.router.rpc import rpc_call as _rpc_call
from tigrbl_concrete._mapping.model import rebind as _rebind, bind as _bind
from ._op_registry import get_registry
from tigrbl_core._spec import OpSpec
from ._table_registry import TableRegistry
from ._routing import include_router as _include_router_impl
from tigrbl_concrete.system import mount_openrpc as _mount_openrpc
from tigrbl_concrete.system import mount_diagnostics as _mount_diagnostics
from tigrbl_concrete.system.docs import build_openapi as _build_openapi
from tigrbl_concrete._concrete import engine_resolver as _resolver
from ._engine import Engine


class TigrblRouter(_Router):
    """
    Canonical router-focused facade that owns:
      • containers (tables, schemas, handlers, hooks, rpc, rest, routers, columns, table_config, core proxies)
      • table inclusion (REST + RPC wiring)
      • (optional) auth knobs recognized by some middlewares/dispatchers

    It composes v3 primitives; you can still use the functions directly if you prefer.
    """

    PREFIX = ""
    REST_PREFIX = "/api"
    RPC_PREFIX = "/rpc"
    SYSTEM_PREFIX = "/system"
    TAGS: Sequence[Any] = ()
    ROUTERS: Sequence[Any] = ()
    TABLES: Sequence[Any] = ()

    # --- optional auth knobs recognized by some middlewares/dispatchers (kept for back-compat) ---
    _authn: Any = None
    _allow_anon: bool = True
    _authorize: Any = None
    _optional_authn_dep: Any = None
    _allow_anon_ops: set[str] = set()

    @staticmethod
    def _collect_declared_tables(owner: type) -> tuple[Any, ...]:
        collected: list[Any] = []
        seen: set[int] = set()
        for base in owner.__mro__:
            if "TABLES" not in base.__dict__:
                continue
            for table in tuple(base.__dict__.get("TABLES", ()) or ()):
                marker = id(table)
                if marker in seen:
                    continue
                seen.add(marker)
                collected.append(table)
        return tuple(collected)

    def __init__(
        self,
        *,
        engine: EngineCfg | None = None,
        tables: Sequence[type] | None = None,
        prefix: str | None = None,
        jsonrpc_prefix: str | None = None,
        system_prefix: str | None = None,
        router_hooks: Mapping[str, Iterable[Callable]]
        | Mapping[str, Mapping[str, Iterable[Callable]]]
        | None = None,
        **router_kwargs: Any,
    ) -> None:
        if prefix is not None:
            self.PREFIX = prefix
        _Router.__init__(self, engine=engine, **router_kwargs)
        self._authn = None
        self._allow_anon = True
        self._authorize = None
        self._optional_authn_dep = None
        self._allow_anon_ops: set[str] = set()
        self.jsonrpc_prefix = (
            jsonrpc_prefix
            if jsonrpc_prefix is not None
            else getattr(self, "RPC_PREFIX", getattr(self, "JSONRPC_PREFIX", "/rpc"))
        )
        self.system_prefix = (
            system_prefix
            if system_prefix is not None
            else getattr(self, "SYSTEM_PREFIX", "/system")
        )
        # ``prefix=...`` is the canonical REST mount prefix for router-scoped
        # table routes. Fall back to REST_PREFIX only when no explicit prefix
        # is provided.
        self.rest_prefix = (
            prefix if prefix is not None else getattr(self, "REST_PREFIX", "/api")
        )

        # public containers (mirrors used by bindings.router)
        self.tables = TableRegistry(
            tables=self._collect_declared_tables(self.__class__)
        )
        self.schemas = SimpleNamespace()
        self.handlers = SimpleNamespace()
        self.hooks = SimpleNamespace()
        self.rpc = SimpleNamespace()
        self.rest = SimpleNamespace()
        self.routers: Dict[str, Any] = {}
        self.columns: Dict[str, Tuple[str, ...]] = {}
        self.table_config: Dict[str, Dict[str, Any]] = {}
        self.core = SimpleNamespace()
        self.core_raw = SimpleNamespace()

        # Router-level hooks map (merged into each model at include-time; precedence handled in bindings.hooks)
        self._router_hooks_map = copy.deepcopy(router_hooks) if router_hooks else None
        if tables:
            self.include_tables(list(tables))

    # ------------------------- internal helpers -------------------------

    @staticmethod
    def _merge_router_hooks_into_model(model: type, hooks_map: Any) -> None:
        """
        Install Router-level hooks on the model so the binder can see them.
        Accepted shapes:
            {phase: [fn, ...]}                           # global, all aliases
            {alias: {phase: [fn, ...]}, "*": {...}}      # per-alias + wildcard
        If the model already has __tigrbl_router_hooks__, we shallow-merge keys.
        """
        if not hooks_map:
            return
        existing = getattr(model, "__tigrbl_router_hooks__", None)
        if existing is None:
            setattr(model, "__tigrbl_router_hooks__", copy.deepcopy(hooks_map))
            return

        # shallow merge (alias or phase keys); values are lists we extend
        merged = copy.deepcopy(existing)
        for k, v in (hooks_map or {}).items():
            if k not in merged:
                merged[k] = copy.deepcopy(v)
            else:
                # when both are dicts, merge phase lists
                if isinstance(v, Mapping) and isinstance(merged[k], Mapping):
                    for ph, fns in v.items():
                        merged[k].setdefault(ph, [])
                        merged[k][ph] = list(merged[k][ph]) + list(fns or [])
                else:
                    # fallback: prefer model-local value, then append router-level
                    if isinstance(merged[k], list):
                        merged[k] = list(merged[k]) + list(v or [])
                    else:
                        merged[k] = v
        setattr(model, "__tigrbl_router_hooks__", merged)

    # ------------------------- primary operations -------------------------

    def include_table(
        self, model: type, *, prefix: str | None = None, mount_router: bool = True
    ) -> Tuple[type, Any]:
        """
        Bind a model, mount its REST router, and attach all namespaces to this facade.
        """
        # inject Router-level hooks so the binder merges them
        self._merge_router_hooks_into_model(model, self._router_hooks_map)
        included_table, router = _include_table(
            self, model, app=None, prefix=prefix, mount_router=mount_router
        )
        if mount_router and router is not None:
            _include_router_impl(self, router, prefix=prefix)
        return included_table, router

    def include_tables(
        self,
        tables: Sequence[type],
        *,
        base_prefix: str | None = None,
        mount_router: bool = True,
    ) -> Dict[str, Any]:
        for m in tables:
            self._merge_router_hooks_into_model(m, self._router_hooks_map)
        included = _include_tables(
            self,
            tables,
            app=None,
            base_prefix=base_prefix,
            mount_router=mount_router,
        )
        if mount_router:
            for router in included.values():
                if router is not None:
                    _include_router_impl(self, router, prefix=base_prefix)
        return included

    def install_engines(
        self, *, router: Any | None = None, tables: tuple[Any, ...] | None = None
    ) -> None:
        """Install engine providers for this router and optional table set."""
        selected_router = self if router is None else router
        selected_tables = tables if tables is not None else tuple(self.tables.values())
        Engine.install_from_objects(
            router=selected_router, tables=tuple(selected_tables)
        )

    def initialize(
        self,
        *,
        schemas: Iterable[str] | None = None,
        sqlite_attachments: Mapping[str, str] | None = None,
        tables: Iterable[Any] | None = None,
    ):
        """Initialize DDL for this router."""
        try:
            result = _ddl_initialize(
                self,
                schemas=schemas,
                sqlite_attachments=sqlite_attachments,
                tables=tables,
            )
        except ValueError as exc:
            if str(exc) != "Engine provider is not configured":
                raise
            result = None

        if inspect.isawaitable(result):

            async def _inner() -> None:
                await result

            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                asyncio.run(_inner())
                return None
            return loop.create_task(_inner())

        try:
            asyncio.get_running_loop()
        except RuntimeError:
            return None

        async def _noop() -> None:
            return None

        return _noop()

    async def rpc_call(
        self,
        model_or_name: type | str,
        method: str,
        payload: Any = None,
        *,
        db: Any | None = None,
        request: Any = None,
        ctx: Optional[Dict[str, Any]] = None,
    ) -> Any:
        return await _rpc_call(
            self, model_or_name, method, payload, db=db, request=request, ctx=ctx
        )

    def mount_jsonrpc(
        self, *, prefix: str | None = None, tags: Sequence[str] | None = ("rpc",)
    ) -> Any:
        del tags
        if prefix is not None:
            self.jsonrpc_prefix = prefix

        existing_paths = {getattr(route, "path", None) for route in self.routes}
        if (
            self.jsonrpc_prefix not in existing_paths
            and f"{self.jsonrpc_prefix}/" not in existing_paths
        ):
            self.add_route(
                self.jsonrpc_prefix,
                lambda *_args, **_kwargs: None,
                methods=["POST"],
            )
        return self

    def mount_openrpc(
        self,
        *,
        path: str = "/openrpc.json",
        name: str = "openrpc_json",
        tags: Sequence[str] | None = None,
    ) -> Any:
        """Mount an OpenRPC JSON endpoint onto this router instance."""
        return _mount_openrpc(self, path=path, name=name, tags=tags)

    def attach_diagnostics(
        self, *, prefix: str | None = None, app: Any | None = None
    ) -> Any:
        """Mount diagnostics router onto this router or the provided ``app``."""
        px = prefix if prefix is not None else self.system_prefix
        prov = _resolver.resolve_provider(router=self)
        get_db = prov.get_db if prov else None
        router = _mount_diagnostics(self, get_db=get_db)
        _include_router_impl(self, router, prefix=px)
        if app is not None and app is not self:
            include_other = getattr(app, "include_router", None)
            if callable(include_other):
                include_other(router, prefix=px)
            app_tables = getattr(app, "tables", None)
            if isinstance(app_tables, dict):
                for name, model in getattr(self, "tables", {}).items():
                    app_tables.setdefault(name, model)
                default_router = getattr(app, "_default_router", None)
                if default_router is not None and isinstance(
                    getattr(default_router, "tables", None), dict
                ):
                    for name, model in getattr(self, "tables", {}).items():
                        default_router.tables.setdefault(name, model)
        return router

    def openapi(self) -> Dict[str, Any]:
        """Build and return the OpenAPI document for this router."""
        return _build_openapi(self)

    # ------------------------- registry passthroughs -------------------------

    def registry(self, model: type):
        """Return the per-model OpspecRegistry."""
        return get_registry(model)

    def bind(self, model: type) -> Tuple[OpSpec, ...]:
        """Bind/rebuild a model in place (without mounting)."""
        self._merge_router_hooks_into_model(model, self._router_hooks_map)
        return _bind(model)

    def rebind(
        self, model: type, *, changed_keys: Optional[set[tuple[str, str]]] = None
    ) -> Tuple[OpSpec, ...]:
        """Targeted rebuild of a bound model."""
        return _rebind(model, changed_keys=changed_keys)

    # Optional: let callers set auth knobs used by some middlewares/dispatchers
    def set_auth(
        self,
        *,
        authn: Any = None,
        allow_anon: Optional[bool] = None,
        authorize: Any = None,
        optional_authn_dep: Any = None,
    ) -> None:
        if authn is not None:
            self._authn = authn
            if allow_anon is None:
                allow_anon = False
        if allow_anon is not None:
            self._allow_anon = bool(allow_anon)
        if authorize is not None:
            self._authorize = authorize
        if optional_authn_dep is not None:
            self._optional_authn_dep = optional_authn_dep

        # Refresh already-included models so routers pick up new auth settings
        if self.tables:
            self._refresh_security()

    def _resolve_registered_model(self, name: str, value: Any) -> Any:
        if isinstance(value, type):
            return value
        core_proxy = getattr(self.core, name, None)
        model = getattr(core_proxy, "_model", None)
        return model if isinstance(model, type) else None

    def _refresh_security(self) -> None:
        """Re-seed auth deps on models and rebuild routers."""
        # Reset routes and allow_anon ops cache
        self.routes = []
        self._allow_anon_ops = set()
        for name, registered in self.tables.items():
            model = self._resolve_registered_model(name, registered)
            if not isinstance(model, type):
                continue
            self.include_table(model, mount_router=False)

    def _collect_tables(self):
        # dedupe; handle multiple DeclarativeBases (multiple metadatas)
        seen = set()
        tables = []
        for name, registered in self.tables.items():
            table = (
                registered
                if not isinstance(registered, type)
                and hasattr(registered, "metadata")
                and hasattr(registered, "columns")
                else None
            )
            if table is None:
                model = self._resolve_registered_model(name, registered)
                table = (
                    getattr(model, "__table__", None)
                    if isinstance(model, type)
                    else None
                )
            if table is not None and not table.columns:
                continue
            if table is not None and table not in seen:
                seen.add(table)
                tables.append(table)
        return tables

    # ------------------------- repr -------------------------

    def __repr__(self) -> str:  # pragma: no cover
        tables = list(getattr(self, "tables", {}))
        rpc_ns = getattr(self, "rpc", None)
        rpc_keys = list(getattr(rpc_ns, "__dict__", {}).keys()) if rpc_ns else []
        return f"<TigrblRouter tables={tables} rpc={rpc_keys}>"



"""OpenAPI helpers and mount utilities."""

from .helpers import (
    _extract_security_dependencies,
    _is_http_bearer_dependency,
    _iter_security_dependencies,
    _normalize_schema_refs,
    _request_schema_from_handler,
    _resolve_component_schema_ref,
    _schema_from_annotation,
    _schema_from_model,
    _security_from_dependencies,
    _security_schemes_from_dependencies,
)
from .mount import mount_openapi
from .schema import openapi

build_openapi = openapi

__all__ = [
    "openapi",
    "build_openapi",
    "mount_openapi",
    "_extract_security_dependencies",
    "_is_http_bearer_dependency",
    "_iter_security_dependencies",
    "_normalize_schema_refs",
    "_request_schema_from_handler",
    "_resolve_component_schema_ref",
    "_schema_from_annotation",
    "_schema_from_model",
    "_security_from_dependencies",
    "_security_schemes_from_dependencies",
]



from __future__ import annotations
from typing import Any, Callable, Optional, TypeVar, overload

from tigrbl_core._spec.response_spec import ResponseSpec

T = TypeVar("T")
_ATTR = "__tigrbl_response_spec__"
_ALIAS_ATTR = "__tigrbl_response_alias__"


def _to_spec(spec: Optional[ResponseSpec] = None, **kwargs: Any) -> ResponseSpec:
    if spec is not None and kwargs:
        raise TypeError("response_ctx: provide either a ResponseSpec or keyword args")
    if spec is not None:
        return spec
    return ResponseSpec(**kwargs)


@overload
def response_ctx(spec: ResponseSpec) -> Callable[[T], T]: ...


@overload
def response_ctx(**kwargs: Any) -> Callable[[T], T]: ...


def response_ctx(*args: Any, **kwargs: Any) -> Callable[[T], T]:
    alias = kwargs.pop("alias", None)
    spec = _to_spec(*args, **kwargs)

    def decorator(target: T) -> T:
        setattr(target, _ATTR, spec)
        if alias is not None:
            setattr(target, _ALIAS_ATTR, alias)
        return target

    return decorator


def get_attached_response_spec(obj: Any) -> Optional[ResponseSpec]:
    return getattr(obj, _ATTR, None)


def get_attached_response_alias(obj: Any) -> Optional[str]:
    return getattr(obj, _ALIAS_ATTR, None)



"""HTTPX transport compatibility helpers."""

from __future__ import annotations

import importlib
import importlib.util
from typing import Any


def ensure_httpx_sync_transport() -> None:
    spec = importlib.util.find_spec("httpx")
    if spec is None:
        return
    httpx = importlib.import_module("httpx")
    if hasattr(httpx.ASGITransport, "__enter__"):
        return

    def __enter__(self) -> Any:
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        return None

    httpx.ASGITransport.__enter__ = __enter__
    httpx.ASGITransport.__exit__ = __exit__



"""Engine utilities for collecting and binding database providers."""

from __future__ import annotations

from tigrbl_concrete._concrete._engine import Engine
from tigrbl_concrete.engine.builders import (
    HybridSession,
    async_postgres_engine,
    async_sqlite_engine,
    blocking_postgres_engine,
    blocking_sqlite_engine,
)
from tigrbl_concrete.engine.collect import collect_engine_config
from tigrbl_concrete.engine.plugins import (
    load_engine_plugins as _bootstrap_load_engine_plugins,
)


__all__ = [
    "collect_engine_config",
    "collect",
    "install",
    "bind",
    "install_from_objects",
    "resolver",
    "blocking_sqlite_engine",
    "blocking_postgres_engine",
    "async_sqlite_engine",
    "async_postgres_engine",
    "HybridSession",
    "Engine",
    "EngineSpec",
    "engine",
    "load_engine_plugins",
    "register_engine",
    "known_engine_kinds",
    "get_engine_registration",
]


def collect(*args, **kwargs):
    from tigrbl_concrete.engine.bind import collect as _collect

    return _collect(*args, **kwargs)


def install(*args, **kwargs):
    from tigrbl_concrete.engine.bind import install as _install

    return _install(*args, **kwargs)


def bind(*args, **kwargs):
    from tigrbl_concrete.engine.bind import bind as _bind

    return _bind(*args, **kwargs)


def install_from_objects(*args, **kwargs):
    from tigrbl_concrete.engine.bind import (
        install_from_objects as _install_from_objects,
    )

    return _install_from_objects(*args, **kwargs)


def load_engine_plugins(*args, **kwargs):
    from tigrbl_concrete.engine.plugins import (
        load_engine_plugins as _load_engine_plugins,
    )

    return _load_engine_plugins(*args, **kwargs)


def register_engine(*args, **kwargs):
    from tigrbl_concrete.engine.registry import register_engine as _register_engine

    return _register_engine(*args, **kwargs)


def known_engine_kinds(*args, **kwargs):
    from tigrbl_concrete.engine.registry import (
        known_engine_kinds as _known_engine_kinds,
    )

    return _known_engine_kinds(*args, **kwargs)


def get_engine_registration(*args, **kwargs):
    from tigrbl_concrete.engine.registry import (
        get_engine_registration as _get_engine_registration,
    )

    return _get_engine_registration(*args, **kwargs)


try:
    _bootstrap_load_engine_plugins()
except Exception:
    pass


def engine(*args, **kwargs):
    from tigrbl.shortcuts.engine import engine as _engine

    return _engine(*args, **kwargs)


def _register_concrete_defaults() -> None:
    """Register built-in concrete engine kinds in the shared core registry."""
    from tigrbl_core._spec.registry import register_engine

    class _SQLiteRegistration:
        def build(self, *, mapping, spec, dsn):
            if spec.async_:
                return async_sqlite_engine(path=spec.path, pool=spec.pool)
            return blocking_sqlite_engine(path=spec.path, pool=spec.pool)

        def capabilities(self, *, spec, mapping=None):
            return {
                "transactional": True,
                "async_native": bool(spec.async_),
                "isolation_levels": {"SERIALIZABLE", "READ UNCOMMITTED"},
                "read_only_enforced": False,
                "engine": "sqlite",
            }

    class _PostgresRegistration:
        def build(self, *, mapping, spec, dsn):
            if spec.async_:
                return async_postgres_engine(
                    dsn=dsn or spec.dsn,
                    user=spec.user or "app",
                    pwd=spec.pwd,
                    host=spec.host or "localhost",
                    port=spec.port or 5432,
                    db=spec.name or "app_db",
                    pool_size=spec.pool_size,
                    max_size=spec.max,
                )
            return blocking_postgres_engine(
                dsn=dsn or spec.dsn,
                user=spec.user or "app",
                pwd=spec.pwd,
                host=spec.host or "localhost",
                port=spec.port or 5432,
                db=spec.name or "app_db",
                pool_size=spec.pool_size,
                max_overflow=spec.max,
            )

        def capabilities(self, *, spec, mapping=None):
            return {
                "transactional": True,
                "async_native": bool(spec.async_),
                "isolation_levels": {
                    "READ UNCOMMITTED",
                    "READ COMMITTED",
                    "REPEATABLE READ",
                    "SERIALIZABLE",
                },
                "read_only_enforced": True,
                "engine": "postgres",
            }

    register_engine("sqlite", _SQLiteRegistration())
    register_engine("postgres", _PostgresRegistration())


def __getattr__(name: str):
    if name == "resolver":
        from tigrbl_concrete._concrete import engine_resolver as resolver

        return resolver
    if name == "EngineSpec":
        from tigrbl_core._spec.engine_spec import EngineSpec

        return EngineSpec
    raise AttributeError(name)



"""Engine plugin loading and built-in concrete registrations."""

from __future__ import annotations

from tigrbl_core._spec.plugins import load_engine_plugins


def register() -> None:
    """Backward-compatible registration hook for concrete engines."""
    from . import _register_concrete_defaults

    _register_concrete_defaults()


__all__ = ["load_engine_plugins", "register"]



"""Dependency wrappers for API/security injection semantics."""

from __future__ import annotations

from typing import Any, Callable


class Dependency:
    def __init__(self, dependency: Callable[..., Any]) -> None:
        self.dependency = dependency


def Depends(fn: Callable[..., Any]) -> Dependency:
    return Dependency(fn)


def Security(fn: Callable[..., Any]) -> Dependency:
    return Dependency(fn)


__all__ = ["Dependency", "Depends", "Security"]



from __future__ import annotations

import inspect
from typing import Any, Callable, Optional

from tigrbl_base._base import TigrblSessionBase
from tigrbl_core._spec.session_spec import SessionSpec


class DefaultSession(TigrblSessionBase):
    """Delegating wrapper for native provider sessions."""

    def __init__(self, underlying: Any, spec: Optional[SessionSpec] = None) -> None:
        super().__init__(spec)
        self._u = underlying

    async def run_sync(self, fn: Callable[[Any], Any]) -> Any:
        rv = fn(self._u)
        if inspect.isawaitable(rv):
            return await rv
        return rv

    async def _tx_begin_impl(self) -> None:
        fn = getattr(self._u, "begin", None)
        if not callable(fn):
            raise RuntimeError("underlying session does not support begin()")
        rv = fn()
        if inspect.isawaitable(rv):
            await rv

    async def _tx_commit_impl(self) -> None:
        fn = getattr(self._u, "commit", None)
        if not callable(fn):
            raise RuntimeError("underlying session does not support commit()")
        rv = fn()
        if inspect.isawaitable(rv):
            await rv

    async def _tx_rollback_impl(self) -> None:
        fn = getattr(self._u, "rollback", None)
        if not callable(fn):
            raise RuntimeError("underlying session does not support rollback()")
        rv = fn()
        if inspect.isawaitable(rv):
            await rv

    def in_transaction(self) -> bool:
        it = getattr(self._u, "in_transaction", None)
        if callable(it):
            try:
                return bool(it())
            except Exception:
                pass
        return super().in_transaction()

    def _add_impl(self, obj: Any) -> Any:
        fn = getattr(self._u, "add", None)
        if not callable(fn):
            raise NotImplementedError("underlying session does not implement add(obj)")
        return fn(obj)

    async def _delete_impl(self, obj: Any) -> None:
        fn = getattr(self._u, "delete", None)
        if not callable(fn):
            raise NotImplementedError(
                "underlying session does not implement delete(obj)"
            )
        rv = fn(obj)
        if inspect.isawaitable(rv):
            await rv

    async def _flush_impl(self) -> None:
        fn = getattr(self._u, "flush", None)
        if callable(fn):
            rv = fn()
            if inspect.isawaitable(rv):
                await rv

    async def _refresh_impl(self, obj: Any) -> None:
        fn = getattr(self._u, "refresh", None)
        if callable(fn):
            rv = fn(obj)
            if inspect.isawaitable(rv):
                await rv

    async def _get_impl(self, model: type, ident: Any) -> Any | None:
        fn = getattr(self._u, "get", None)
        if not callable(fn):
            raise NotImplementedError(
                "underlying session does not implement get(model, ident)"
            )
        rv = fn(model, ident)
        return await rv if inspect.isawaitable(rv) else rv

    async def _execute_impl(self, stmt: Any) -> Any:
        fn = getattr(self._u, "execute", None)
        if not callable(fn):
            raise NotImplementedError(
                "underlying session does not implement execute(stmt)"
            )
        rv = fn(stmt)
        return await rv if inspect.isawaitable(rv) else rv

    async def _close_impl(self) -> None:
        fn = getattr(self._u, "close", None)
        if callable(fn):
            rv = fn()
            if inspect.isawaitable(rv):
                await rv


def wrap_sessionmaker(maker: Callable[..., Any], spec: SessionSpec):
    """Wrap a provider session factory to yield DefaultSession instances."""

    def _mk() -> DefaultSession:
        underlying = maker()
        session = DefaultSession(underlying, spec)
        session.apply_spec(spec)
        return session

    return _mk


__all__ = ["DefaultSession", "wrap_sessionmaker"]



"""Shared case-insensitive HTTP header mapping primitives."""

from __future__ import annotations

from collections.abc import Iterable, Mapping, MutableMapping
from typing import Any
from http.cookies import SimpleCookie


class HeaderCookies(dict[str, str]):
    """Dot-addressable cookie mapping parsed from a Cookie header value."""

    def __getattr__(self, name: str) -> str:
        try:
            return self[name]
        except KeyError as exc:  # pragma: no cover - defensive
            raise AttributeError(name) from exc


class SetCookieHeader(str):
    """String-like Set-Cookie header value with cookie-name lookup convenience."""

    def get(self, name: str, default: str | None = None) -> str | None:
        parsed = SimpleCookie()
        parsed.load(str(self))
        morsel = parsed.get(name)
        if morsel is None:
            return default
        return morsel.value


class Headers(MutableMapping[str, str]):
    """Case-insensitive header mapping for request/response objects."""

    def __init__(
        self,
        values: Iterable[tuple[str, str]] | Mapping[str, str] | None = None,
    ) -> None:
        self._data: dict[str, tuple[str, str]] = {}
        if values is None:
            return
        items = values.items() if hasattr(values, "items") else values
        for key, value in items:
            self[key] = self._normalize_value(value)

    @staticmethod
    def _normalize_key(key: Any) -> str:
        if isinstance(key, (bytes, bytearray)):
            return key.decode("latin-1").lower()
        return str(key).lower()

    @staticmethod
    def _normalize_value(value: Any) -> str:
        if isinstance(value, (bytes, bytearray)):
            return value.decode("latin-1")
        return str(value)

    @staticmethod
    def _attribute_key(name: str) -> str:
        return name.replace("_", "-").lower()

    @staticmethod
    def _parse_cookie_header(value: str) -> HeaderCookies:
        parsed = SimpleCookie()
        parsed.load(value)
        return HeaderCookies({name: morsel.value for name, morsel in parsed.items()})

    @staticmethod
    def _coerce_lookup_value(key: str, value: str) -> str | SetCookieHeader:
        if key == "set-cookie":
            return SetCookieHeader(value)
        return value

    def __getitem__(self, key: str) -> str | SetCookieHeader:
        normalized = self._normalize_key(key)
        return self._coerce_lookup_value(normalized, self._data[normalized][1])

    def __setitem__(self, key: str, value: str) -> None:
        lower = self._normalize_key(key)
        normalized_value = self._normalize_value(value)
        if lower == "set-cookie" and lower in self._data:
            existing = self._data[lower][1]
            merged = f"{existing}, {normalized_value}" if existing else normalized_value
            self._data[lower] = (lower, merged)
            return
        self._data[lower] = (lower, normalized_value)

    def __delitem__(self, key: str) -> None:
        del self._data[self._normalize_key(key)]

    def __iter__(self):
        for original, _ in self._data.values():
            yield original

    def __len__(self) -> int:
        return len(self._data)

    def items(self):
        return ((k, v) for k, v in self._data.values())

    def as_list(self) -> list[tuple[str, str]]:
        return list(self.items())

    def append(self, item: tuple[str, str]) -> None:
        key, value = item
        self[key] = value

    def __contains__(self, item: object) -> bool:
        if isinstance(item, tuple) and len(item) == 2:
            key, value = item
            if not isinstance(key, str):
                return False
            current = self.get(key)
            return current == str(value) if current is not None else False
        if isinstance(item, str):
            return self.get(item) is not None
        return False

    def get(self, key: str, default: str | None = None) -> str | SetCookieHeader | None:
        normalized = self._normalize_key(key)
        item = self._data.get(normalized)
        if item is None:
            return default
        return self._coerce_lookup_value(normalized, item[1])

    def __getattr__(self, name: str) -> str | HeaderCookies:
        key = self._attribute_key(name)
        if key not in self._data:
            raise AttributeError(name)
        value = self._data[key][1]
        if key == "cookie":
            return self._parse_cookie_header(value)
        return value

    def __setattr__(self, name: str, value: str) -> None:
        if name.startswith("_"):
            object.__setattr__(self, name, value)
            return
        self[self._attribute_key(name)] = value


__all__ = ["Headers", "HeaderCookies", "SetCookieHeader"]



"""Operation-related decorators for Tigrbl v3."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Iterable, Optional, Sequence, Union

from tigrbl_core._spec.op_spec import OpSpec, Arity, TargetOp, PersistPolicy
from tigrbl_core._spec.op_utils import (
    _maybe_await as _core_maybe_await,
    _normalize_persist as _core_normalize_persist,
    _unwrap as _core_unwrap,
)
from tigrbl_core._spec.schema_spec import SchemaArg


# ---------------------------------------------------------------------------
# Utilities
# ---------------------------------------------------------------------------


def _unwrap(obj: Any) -> Callable[..., Any]:
    """Get underlying function for (class|static)method; else return obj."""
    return _core_unwrap(obj)


def _ensure_cm(func: Any) -> Any:
    """Ensure method is a classmethod so it receives (cls, ctx)."""
    if isinstance(func, (classmethod, staticmethod)):
        return func
    return classmethod(func)


def _maybe_await(v):
    return _core_maybe_await(v)


# ---------------------------------------------------------------------------
# alias_ctx with optional rich overrides
# ---------------------------------------------------------------------------


@dataclass(frozen=True)
class AliasDecl:
    alias: str
    # Optional overrides (lazy-capable schema args are fine; resolved later)
    request_schema: Optional[SchemaArg] = None
    response_schema: Optional[SchemaArg] = None
    persist: Optional[PersistPolicy] = None
    arity: Optional[Arity] = None
    rest: Optional[bool] = None


def alias(name: str, **kw) -> AliasDecl:
    """Convenience helper: alias('get', response_schema=..., rest=False)."""
    return AliasDecl(alias=name, **kw)


def alias_ctx(**verb_to_alias_or_decl: Union[str, AliasDecl]):
    """Class decorator mapping canonical verbs → aliases with optional overrides."""

    def deco(cls: type):
        amap = dict(getattr(cls, "__tigrbl_aliases__", {}) or {})
        overrides = dict(getattr(cls, "__tigrbl_alias_overrides__", {}) or {})

        for canon, value in verb_to_alias_or_decl.items():
            if isinstance(value, AliasDecl):
                amap[canon] = value.alias
                overrides[canon] = {
                    "request_schema": value.request_schema,
                    "response_schema": value.response_schema,
                    "persist": value.persist,
                    "arity": value.arity,
                    "rest": value.rest,
                }
            elif isinstance(value, str):
                amap[canon] = value
            else:
                raise TypeError(
                    f"alias_ctx[{canon}] must be str or AliasDecl, got {type(value)}"
                )

        setattr(cls, "__tigrbl_aliases__", amap)
        setattr(cls, "__tigrbl_alias_overrides__", overrides)
        # Alias maps are rebuilt from class metadata at bind-time.
        return cls

    return deco


# ---------------------------------------------------------------------------
# op_alias (class decorator): attach an OpSpec alias to the model
# ---------------------------------------------------------------------------


def op_alias(
    *,
    alias: str,
    target: TargetOp,
    arity: Arity | None = None,
    persist: PersistPolicy = "default",
    request_model: SchemaArg | None = None,
    response_model: SchemaArg | None = None,
    http_methods: Sequence[str] | None = None,
    tags: Sequence[str] | None = None,
):
    """Class decorator to declare an alias for an operation."""

    def deco(table_cls: type):
        ops = list(getattr(table_cls, "__tigrbl_ops__", ()))
        inferred_arity = arity or _infer_arity(target)
        if arity is None and target == "custom" and table_cls.__name__.endswith("App"):
            inferred_arity = "collection"

        spec = OpSpec(
            alias=alias,
            target=target,
            table=table_cls,
            arity=inferred_arity,
            persist=_normalize_persist(persist),
            request_model=request_model,
            response_model=response_model,
            http_methods=tuple(http_methods) if http_methods else None,
            tags=tuple(tags or ()),
        )
        ops.append(spec)
        table_cls.__tigrbl_ops__ = tuple(ops)
        return table_cls

    return deco


# ---------------------------------------------------------------------------
# op_ctx (single path: target + arity) with schema overrides
# ---------------------------------------------------------------------------


def op_ctx(
    *,
    bind: Any | Iterable[Any] | None = None,
    alias: Optional[str] = None,
    target: Optional[TargetOp] = None,
    arity: Optional[Arity] = None,
    http_methods: Sequence[str] | None = None,
    rest: Optional[bool] = None,
    request_schema: Optional[SchemaArg] = None,
    response_schema: Optional[SchemaArg] = None,
    persist: Optional[PersistPolicy] = None,
    status_code: Optional[int] = None,
):
    """Declare a ctx-only operation whose body is `(cls, ctx)`."""

    def deco(fn):
        cm = _ensure_cm(fn)
        f = _unwrap(cm)
        f.__tigrbl_ctx_only__ = True
        resolved_target = target or "custom"
        inferred_arity = arity or _infer_arity(resolved_target)
        if (
            arity is None
            and resolved_target == "custom"
            and bind is not None
            and any(
                isinstance(obj, type) and obj.__name__.endswith("App")
                for obj in (
                    bind
                    if isinstance(bind, Iterable) and not isinstance(bind, (str, bytes))
                    else [bind]
                )
            )
        ):
            inferred_arity = "collection"

        if alias is not None:
            resolved_alias = alias
        elif resolved_target != "custom":
            resolved_alias = resolved_target
        elif bind is not None:
            resolved_alias = f.__name__
        else:
            resolved_alias = "custom"

        spec = OpSpec(
            alias=resolved_alias,
            target=resolved_target,
            arity=inferred_arity,
            http_methods=tuple(http_methods) if http_methods else None,
            expose_routes=True if rest is None else bool(rest),
            request_model=request_schema,
            response_model=response_schema,
            persist=_normalize_persist(persist),
            status_code=status_code,
        )
        f.__tigrbl_op_spec__ = spec
        # Backward-compatible attr name; value is always OpSpec.
        f.__tigrbl_op_decl__ = spec

        if bind is not None:
            targets = (
                bind
                if isinstance(bind, Iterable) and not isinstance(bind, (str, bytes))
                else [bind]
            )
            for obj in targets:
                setattr(obj, f.__name__, cm)
                if isinstance(obj, type):
                    _refresh_bound_ops(obj)

        return cm

    return deco


# ---------------------------------------------------------------------------
# Collection helpers
# ---------------------------------------------------------------------------

_COLLECTION_VERBS = {
    "create",
    "list",
    "bulk_create",
    "bulk_update",
    "bulk_replace",
    "bulk_merge",
    "bulk_delete",
    "clear",
    "merge",
}


def _infer_arity(target: str) -> str:
    if target == "custom":
        return "collection"
    return "collection" if target in _COLLECTION_VERBS else "member"


def _normalize_persist(p) -> str:
    return _core_normalize_persist(p)


def _refresh_bound_ops(model: type) -> None:
    try:
        from tigrbl_core._spec.op_spec import _mro_collect_decorated_ops
        from tigrbl_concrete._mapping.model import (
            _bind_model_hooks,
            _materialize_handlers,
        )

        specs = tuple(_mro_collect_decorated_ops(model))
        _materialize_handlers(model, specs)
        _bind_model_hooks(model, specs)
    except Exception:
        # Best-effort refresh for dynamic binds.
        return


__all__ = ["alias", "alias_ctx", "op_alias", "op_ctx"]



# tigrbl/_concrete/_engine.py
from __future__ import annotations

from contextlib import contextmanager, asynccontextmanager
from dataclasses import dataclass, field
import asyncio
import inspect
from typing import Any, Callable, Optional, Tuple, Union, Protocol, TYPE_CHECKING

from tigrbl_base._base._engine_base import EngineBase
from tigrbl_base._base._engine_provider_base import EngineProviderBase

try:
    from sqlalchemy.orm import Session
    from sqlalchemy.ext.asyncio import AsyncSession
except Exception:  # pragma: no cover
    Session = object  # type: ignore
    AsyncSession = object  # type: ignore


class SessionFactory(Protocol):
    def __call__(self) -> Union[Session, AsyncSession]: ...


Builder = Callable[[], Tuple[Any, SessionFactory]]  # returns (engine, sessionmaker)

if TYPE_CHECKING:  # pragma: no cover - for type checkers only
    from tigrbl_core._spec.engine_spec import EngineProviderSpec, EngineSpec


@dataclass(frozen=True)
class EngineProvider(EngineProviderBase):
    # supports() exposes engine capabilities for compatibility checks
    def supports(self) -> dict:
        try:
            return self.spec.supports()
        except Exception:
            return {"engine": self.spec.kind or "unknown"}

    """Lazily builds an engine + sessionmaker from an :class:`EngineSpec`."""

    spec: "EngineSpec"
    _engine: Any = None
    _maker: Optional[SessionFactory] = None
    get_db: Callable[..., Any] = field(init=False)

    @property
    def kind(self) -> str:
        return "async" if self.spec.async_ else "sync"

    def to_provider(self) -> "EngineProviderBase":
        return self

    def ensure(self) -> Tuple[Any, SessionFactory]:
        if self._maker is None:
            eng, mk = self._build_with_fallbacks()
            object.__setattr__(self, "_engine", eng)
            object.__setattr__(self, "_maker", mk)
        return self._engine, self._maker  # type: ignore[return-value]

    def _build_with_fallbacks(self) -> Tuple[Any, SessionFactory]:
        try:
            return self.spec.build()
        except RuntimeError as exc:
            if "Unknown or unavailable engine kind" not in str(exc):
                raise

        # Workspace/dev runs may not expose plugin entry-point metadata.
        try:
            from tigrbl_concrete.engine.plugins import register as _register_concrete

            _register_concrete()
            return self.spec.build()
        except Exception:
            pass

        # Legacy local fallback for concrete sqlite/postgres builders.
        from tigrbl_concrete.engine import builders as _builders

        if self.spec.kind == "sqlite":
            if self.spec.async_:
                return _builders.async_sqlite_engine(
                    path=self.spec.path,
                    pool=self.spec.pool,
                )
            return _builders.blocking_sqlite_engine(
                path=self.spec.path,
                pool=self.spec.pool,
            )

        if self.spec.kind == "postgres":
            if self.spec.async_:
                return _builders.async_postgres_engine(
                    user=self.spec.user,
                    pwd=self.spec.pwd,
                    host=self.spec.host,
                    port=self.spec.port,
                    db=self.spec.name,
                    pool_size=self.spec.pool_size,
                    max_size=self.spec.max,
                )
            return _builders.blocking_postgres_engine(
                user=self.spec.user,
                pwd=self.spec.pwd,
                host=self.spec.host,
                port=self.spec.port,
                db=self.spec.name,
                pool_size=self.spec.pool_size,
                max_overflow=self.spec.max,
            )

        return self.spec.build()

    def session(self) -> Union[Session, AsyncSession]:
        _, mk = self.ensure()
        return mk()  # type: ignore[misc]

    def __post_init__(self) -> None:
        if self.spec.async_:

            async def _get_db() -> Any:
                db = self.session()
                try:
                    yield db  # type: ignore[misc]
                finally:
                    close = getattr(db, "close", None)
                    if callable(close):
                        rv = close()
                        if inspect.isawaitable(rv):
                            await rv

            object.__setattr__(self, "get_db", _get_db)
        else:

            def _get_db() -> Any:
                db = self.session()
                try:
                    yield db  # type: ignore[misc]
                finally:
                    close = getattr(db, "close", None)
                    if callable(close):
                        rv = close()
                        if inspect.isawaitable(rv):
                            try:
                                loop = asyncio.get_running_loop()
                            except RuntimeError:
                                asyncio.run(rv)
                            else:
                                loop.create_task(rv)

            object.__setattr__(self, "get_db", _get_db)


@dataclass
class Engine(EngineBase):
    # Delegate to provider/spec for capability reporting
    def supports(self) -> dict:
        try:
            return self.provider.supports()
        except Exception:
            return {"engine": self.spec.kind or "unknown"}

    """Thin façade over an :class:`EngineSpec` with convenient (a)context managers."""

    spec: "EngineSpec"
    provider: EngineProvider = field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "provider", EngineProvider(self.spec))

    @property
    def is_async(self) -> bool:
        return self.provider.kind == "async"

    def to_provider(self) -> "EngineProviderBase":
        return self.provider

    def raw(self) -> Tuple[Any, SessionFactory]:
        return self.provider.ensure()

    @staticmethod
    def collect(
        *,
        app: Any | None = None,
        router: Any | None = None,
        tables: tuple[Any, ...] = (),
    ) -> dict[str, Any]:
        from tigrbl_core.config.engine_traversal import collect_engine_bindings

        return collect_engine_bindings(app=app, router=router, tables=tables)

    @staticmethod
    def install(collected: dict[str, Any]) -> None:
        from tigrbl_concrete._concrete import engine_resolver as _resolver

        default_db = collected.get("default")
        if default_db is not None:
            _resolver.set_default(default_db)
        for router_obj, db in (collected.get("router") or {}).items():
            _resolver.register_router(router_obj, db)
        for table_obj, db in (collected.get("tables") or {}).items():
            _resolver.register_table(table_obj, db)
        for (model, alias), db in (collected.get("ops") or {}).items():
            _resolver.register_op(model, alias, db)

    @staticmethod
    def install_from_objects(
        *,
        app: Any | None = None,
        router: Any | None = None,
        tables: tuple[Any, ...] = (),
    ) -> dict[str, Any]:
        collected = Engine.collect(app=app, router=router, tables=tables)
        Engine.install(collected)
        return collected

    @staticmethod
    def collect_bindings(
        *,
        app: Any | None = None,
        router: Any | None = None,
        tables: tuple[Any, ...] = (),
    ) -> dict[str, Any]:
        """Backward-compatible alias for :meth:`collect`."""
        return Engine.collect(app=app, router=router, tables=tables)

    @staticmethod
    def install_bindings(collected: dict[str, Any]) -> None:
        """Backward-compatible alias for :meth:`install`."""
        Engine.install(collected)

    @property
    def get_db(self) -> Callable[..., Any]:
        return self.provider.get_db

    @contextmanager
    def session(self) -> Session:
        db = self.provider.session()
        try:
            yield db  # type: ignore[return-value]
        finally:
            close = getattr(db, "close", None)
            if callable(close):
                close()

    @asynccontextmanager
    async def asession(self) -> AsyncSession:
        db = self.provider.session()
        try:
            yield db  # type: ignore[return-value]
        finally:
            close = getattr(db, "close", None)
            if callable(close):
                # AsyncSession.close() is sync; close() may exist as async in some impls
                res = close()
                if hasattr(res, "__await__"):
                    await res


def provider_from_spec(spec: "EngineSpec | EngineProviderSpec") -> "EngineProvider":
    """Materialize a concrete :class:`EngineProvider` from a core provider/spec value."""
    if hasattr(spec, "spec"):
        return EngineProvider(spec.spec)  # type: ignore[arg-type]
    return EngineProvider(spec)


Provider = EngineProvider



from __future__ import annotations

from typing import Any, Sequence, Type

from tigrbl_core._spec.app_spec import AppSpec
from tigrbl_concrete._concrete.tigrbl_app import TigrblApp


def defineAppSpec(
    *,
    title: str = "Tigrbl",
    version: str = "0.1.0",
    engine: Any = None,
    routers: Sequence[Any] = (),
    ops: Sequence[Any] = (),
    tables: Sequence[Any] = (),
    schemas: Sequence[Any] = (),
    hooks: Sequence[Any] = (),
    security_deps: Sequence[Any] = (),
    deps: Sequence[Any] = (),
    jsonrpc_prefix: str = "/rpc",
    system_prefix: str = "/system",
    middlewares: Sequence[Any] = (),
    lifespan: Any = None,
) -> Type[AppSpec]:
    attrs = dict(
        TITLE=title,
        VERSION=version,
        ENGINE=engine,
        ROUTERS=tuple(routers or ()),
        OPS=tuple(ops or ()),
        TABLES=tuple(tables or ()),
        SCHEMAS=tuple(schemas or ()),
        HOOKS=tuple(hooks or ()),
        SECURITY_DEPS=tuple(security_deps or ()),
        DEPS=tuple(deps or ()),
        JSONRPC_PREFIX=jsonrpc_prefix,
        SYSTEM_PREFIX=system_prefix,
        MIDDLEWARES=tuple(middlewares or ()),
        LIFESPAN=lifespan,
    )
    return type("AppSpec", (AppSpec,), attrs)


def deriveApp(**kw: Any) -> Type[TigrblApp]:
    """Produce a concrete :class:`TigrblApp` subclass that inherits the spec."""
    spec = defineAppSpec(**kw)
    attrs = {
        "TITLE": getattr(spec, "TITLE", "Tigrbl"),
        "VERSION": getattr(spec, "VERSION", "0.1.0"),
        "ENGINE": getattr(spec, "ENGINE", None),
        "ROUTERS": tuple(getattr(spec, "ROUTERS", ()) or ()),
        "OPS": tuple(getattr(spec, "OPS", ()) or ()),
        "TABLES": tuple(getattr(spec, "TABLES", ()) or ()),
        "SCHEMAS": tuple(getattr(spec, "SCHEMAS", ()) or ()),
        "HOOKS": tuple(getattr(spec, "HOOKS", ()) or ()),
        "SECURITY_DEPS": tuple(getattr(spec, "SECURITY_DEPS", ()) or ()),
        "DEPS": tuple(getattr(spec, "DEPS", ()) or ()),
        "JSONRPC_PREFIX": getattr(spec, "JSONRPC_PREFIX", "/rpc"),
        "SYSTEM_PREFIX": getattr(spec, "SYSTEM_PREFIX", "/system"),
        "MIDDLEWARES": tuple(getattr(spec, "MIDDLEWARES", ()) or ()),
        "LIFESPAN": getattr(spec, "LIFESPAN", None),
    }
    return type("AppWithSpec", (TigrblApp, spec), attrs)


__all__ = ["defineAppSpec", "deriveApp"]



from __future__ import annotations

from typing import Any, Dict, List, Optional
from types import SimpleNamespace

from tigrbl_atoms import HookPhases


def build_hookz_endpoint(router: Any):
    cache: Optional[Dict[str, Dict[str, Dict[str, List[str]]]]] = None

    async def _hookz():
        nonlocal cache
        """
        Expose hook execution order for each method.

        Phases appear in runner order; error phases trail.
        Within each phase, hooks are listed in execution order: global (None) hooks,
        then method-specific hooks.
        """
        if cache is not None:
            return cache

        from . import _table_iter, _opspecs, _label_callable

        out: Dict[str, Dict[str, Dict[str, List[str]]]] = {}
        for model in _table_iter(router):
            mname = getattr(model, "__name__", "Model")
            hooks_root = getattr(model, "hooks", SimpleNamespace())
            alias_sources = set()
            rpc_ns = getattr(model, "rpc", SimpleNamespace())
            alias_sources.update(getattr(rpc_ns, "__dict__", {}).keys())
            for sp in _opspecs(model):
                alias_sources.add(sp.alias)

            model_map: Dict[str, Dict[str, List[str]]] = {}
            for alias in sorted(alias_sources):
                alias_ns = getattr(hooks_root, alias, None) or SimpleNamespace()
                phase_map: Dict[str, List[str]] = {}
                for ph in HookPhases:
                    steps = list(getattr(alias_ns, ph.value, []) or [])
                    if steps:
                        phase_map[ph.value] = [_label_callable(fn) for fn in steps]
                if phase_map:
                    model_map[alias] = phase_map
            out[mname] = model_map
        cache = out
        return cache

    return _hookz



"""Concrete storage-layer primitives."""

from dataclasses import dataclass

from tigrbl_base._base._storage import ForeignKeyBase, StorageTransformBase


@dataclass(frozen=True)
class ForeignKey(ForeignKeyBase):
    """Concrete foreign key configuration used at runtime."""


@dataclass(frozen=True)
class StorageTransform(StorageTransformBase):
    """Concrete storage transform configuration used at runtime."""


__all__ = ["ForeignKey", "StorageTransform"]



"""Compatibility wrapper for engine collection APIs.

``tigrbl_canon`` is deprecated. Keep the public ``collect_engine_config``
entrypoint by delegating to the first-class engine binder helpers.
"""

from __future__ import annotations

from tigrbl_concrete.engine.bind import collect_engine_bindings as collect_engine_config

__all__ = ["collect_engine_config"]



from __future__ import annotations

from typing import Any, Literal
from uuid import UUID, uuid4

from pydantic import BaseModel, Field


def _uuid_examples(schema: dict[str, Any]) -> None:
    """Populate schema examples with a random UUID."""
    schema["examples"] = [str(uuid4())]


class RPCRequest(BaseModel):
    """JSON-RPC 2.0 request envelope."""

    jsonrpc: Literal["2.0"] = "2.0"
    method: str
    params: dict[str, Any] | list[Any] = Field(default_factory=dict)
    id: UUID | str | int | None = Field(
        default_factory=uuid4,
        json_schema_extra=_uuid_examples,
    )


class RPCError(BaseModel):
    code: int
    message: str
    data: Any | None = None


class RPCResponse(BaseModel):
    """JSON-RPC 2.0 response envelope."""

    jsonrpc: Literal["2.0"] = "2.0"
    result: Any | None = None
    error: RPCError | None = None
    id: UUID | str | int | None = Field(
        default=None,
        json_schema_extra=_uuid_examples,
    )



"""Helper utilities for std OpenAPI schema generation."""

from __future__ import annotations

import inspect
from typing import Any, Iterable, get_args, get_origin

from tigrbl_ops_oltp.crud.params import Param
from tigrbl_concrete._mapping.core_resolver import (
    annotation_marker as _annotation_marker,
    split_annotated as _split_annotated,
)
from tigrbl_concrete._concrete._route import Route


def _normalize_schema_refs(node: Any) -> Any:
    if isinstance(node, dict):
        normalized: dict[str, Any] = {}
        for key, value in node.items():
            if (
                key == "$ref"
                and isinstance(value, str)
                and value.startswith("#/$defs/")
            ):
                normalized[key] = value.replace("#/$defs/", "#/components/schemas/")
                continue
            normalized[key] = _normalize_schema_refs(value)
        return normalized
    if isinstance(node, list):
        return [_normalize_schema_refs(item) for item in node]
    return node


def _schema_from_model(
    model: Any, components_schemas: dict[str, Any] | None = None
) -> dict[str, Any]:
    try:
        import pydantic

        if isinstance(model, type) and issubclass(model, pydantic.BaseModel):
            schema = model.model_json_schema(
                ref_template="#/components/schemas/{model}"
            )
            schema = _normalize_schema_refs(schema)
            defs = schema.pop("$defs", None)
            if isinstance(defs, dict) and components_schemas is not None:
                for name, definition in defs.items():
                    components_schemas.setdefault(
                        name, _normalize_schema_refs(definition)
                    )
            if components_schemas is None:
                return schema  # type: ignore[no-any-return]

            schema_name = str(schema.get("title") or model.__name__)
            components_schemas.setdefault(schema_name, schema)
            return {"$ref": f"#/components/schemas/{schema_name}"}
    except Exception:
        pass

    origin = getattr(model, "__origin__", None)
    if origin in (list, tuple, set):
        item = model.__args__[0] if getattr(model, "__args__", None) else Any
        return {"type": "array", "items": _schema_from_model(item, components_schemas)}
    return {"type": "object"}


def _resolve_component_schema_ref(
    schema: dict[str, Any],
    components_schemas: dict[str, Any],
) -> dict[str, Any]:
    ref = schema.get("$ref")
    if not isinstance(ref, str) or not ref.startswith("#/components/schemas/"):
        return schema

    schema_name = ref.rsplit("/", maxsplit=1)[-1]
    component = components_schemas.get(schema_name)
    if not isinstance(component, dict):
        return schema
    return component


def _schema_from_annotation(
    annotation: Any, components_schemas: dict[str, Any]
) -> dict[str, Any]:
    annotation, _ = _split_annotated(annotation)

    if annotation is Any:
        return {"type": "object"}

    origin = get_origin(annotation)
    if origin in (list, tuple, set):
        args = get_args(annotation)
        item_annotation = args[0] if args else Any
        return {
            "type": "array",
            "items": _schema_from_annotation(item_annotation, components_schemas),
        }

    if origin is not None:
        args = [arg for arg in get_args(annotation) if arg is not type(None)]
        if len(args) == 1:
            return _schema_from_annotation(args[0], components_schemas)

    if annotation in (dict,):
        return {"type": "object"}

    try:
        from collections.abc import Mapping as _Mapping

        if isinstance(annotation, type) and issubclass(annotation, _Mapping):
            return {"type": "object"}
    except Exception:
        pass

    return _schema_from_model(annotation, components_schemas)


def _request_schema_from_handler(
    route: Route, components_schemas: dict[str, Any]
) -> dict[str, Any] | None:
    handler = getattr(route, "handler", None)
    if handler is None:
        return None

    try:
        signature = inspect.signature(handler)
    except Exception:
        return None

    for param in signature.parameters.values():
        if param.kind not in (
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
        ):
            continue

        annotation, extras = _split_annotated(param.annotation)
        marker = _annotation_marker(extras, Param)
        if marker is None or marker.location != "body":
            continue

        return _schema_from_annotation(annotation, components_schemas)

    return None


def _security_from_dependencies(deps: Iterable[Any]) -> list[dict[str, list[str]]]:
    security: list[dict[str, list[str]]] = []
    for dependency in _extract_security_dependencies(deps):
        security_dependency = _as_openapi_security_dependency(dependency)
        if security_dependency is None:
            continue
        security.append(security_dependency.openapi_security_requirement())
    return security


def _security_schemes_from_dependencies(deps: Iterable[Any]) -> dict[str, Any]:
    schemes: dict[str, Any] = {}
    for dependency in _extract_security_dependencies(deps):
        security_dependency = _as_openapi_security_dependency(dependency)
        if security_dependency is None:
            continue

        scheme_name = getattr(security_dependency, "scheme_name", None)
        if not isinstance(scheme_name, str) or not scheme_name:
            continue
        schemes[scheme_name] = security_dependency.openapi_security_scheme()
    return schemes


def _extract_security_dependencies(deps: Iterable[Any]) -> Iterable[Any]:
    seen: set[int] = set()
    for dep in deps:
        yield from _iter_security_dependencies(dep, seen)


def _iter_security_dependencies(dep: Any, seen: set[int]) -> Iterable[Any]:
    if dep is None:
        return

    dep_id = id(dep)
    if dep_id in seen:
        return
    seen.add(dep_id)

    dependency = getattr(dep, "dependency", None)
    if dependency is not None:
        yield from _iter_security_dependencies(dependency, seen)
        return

    yield dep

    if not callable(dep):
        return

    try:
        signature = inspect.signature(dep)
    except (TypeError, ValueError):
        return

    for param in signature.parameters.values():
        default = param.default
        if default is inspect._empty:
            continue
        yield from _iter_security_dependencies(default, seen)


def _as_openapi_security_dependency(dep: Any) -> Any | None:
    has_scheme = callable(getattr(dep, "openapi_security_scheme", None))
    has_requirement = callable(getattr(dep, "openapi_security_requirement", None))
    if has_scheme and has_requirement:
        return dep
    return None


def _is_http_bearer_dependency(dep: Any) -> bool:
    security_dependency = _as_openapi_security_dependency(dep)
    if security_dependency is None:
        return False

    scheme = security_dependency.openapi_security_scheme()
    return scheme.get("type") == "http" and scheme.get("scheme") == "bearer"



from __future__ import annotations

from dataclasses import replace
from types import SimpleNamespace
from typing import Any, Callable, Optional

from tigrbl_core._spec.binding_spec import HttpRestBindingSpec
from tigrbl_core._spec.op_spec import OpSpec
from tigrbl_concrete._mapping.model_helpers import _OpSpecGroup
from tigrbl_concrete._concrete._router import Router
from .healthz import build_healthz_endpoint
from .hookz import build_hookz_endpoint
from .kernelz import build_kernelz_endpoint
from .methodz import build_methodz_endpoint
from .utils import maybe_execute, opspecs, table_iter


def _register_runtime_diagnostics_op(
    router: Any,
    *,
    path: str,
    alias: str,
    endpoint_factory: Callable[[Any], Any],
) -> None:
    """Expose diagnostics endpoints to the kernel runtime operation plan."""
    tables = getattr(router, "tables", None)
    if not isinstance(tables, dict):
        return

    model_name = "__tigrbl_system_docs__"
    model = tables.get(model_name)
    if model is None:
        model = type("TigrblSystemDocs", (), {})
        model.resource_name = "system_docs"
        model.hooks = SimpleNamespace()
        model.ops = SimpleNamespace(by_alias={})
        model.opspecs = SimpleNamespace(all=())
        tables[model_name] = model

    binding = HttpRestBindingSpec(proto="http.rest", path=path, methods=("GET",))
    existing_specs = tuple(model.ops.by_alias.get(alias, ()) or ())
    existing = existing_specs[0] if existing_specs else None

    if existing is not None:
        merged_bindings = list(tuple(getattr(existing, "bindings", ()) or ()))
        if binding not in merged_bindings:
            merged_bindings.append(binding)
        op = replace(existing, bindings=tuple(merged_bindings))
    else:
        op = OpSpec(
            alias=alias,
            target="custom",
            arity="collection",
            persist="skip",
            expose_routes=False,
            bindings=(binding,),
        )

    model.ops.by_alias[alias] = _OpSpecGroup((op,))
    model.opspecs.all = tuple(
        spec for spec in model.opspecs.all if getattr(spec, "alias", None) != alias
    ) + (op,)

    async def _runtime_diagnostics_step(ctx: Any) -> None:
        payload = await endpoint_factory(ctx)
        setattr(ctx, "result", payload)
        if isinstance(getattr(ctx, "temp", None), dict):
            ctx.temp.setdefault("egress", {})["result"] = payload

    hooks_ns = getattr(model.hooks, alias, None)
    if hooks_ns is None:
        hooks_ns = SimpleNamespace()
        setattr(model.hooks, alias, hooks_ns)
    hooks_ns.HANDLER = [_runtime_diagnostics_step]


def _register_runtime_paths(
    router: Any, *, path: str, alias: str, endpoint_factory: Callable[[Any], Any]
) -> None:
    _register_runtime_diagnostics_op(
        router,
        path=path,
        alias=alias,
        endpoint_factory=endpoint_factory,
    )
    system_prefix = str(getattr(router, "system_prefix", "/system") or "/system")
    if system_prefix and system_prefix != "/":
        prefixed = f"{system_prefix.rstrip('/')}{path}"
        _register_runtime_diagnostics_op(
            router,
            path=prefixed,
            alias=alias,
            endpoint_factory=endpoint_factory,
        )


def mount_diagnostics(
    router: Any,
    *,
    get_db: Optional[Callable[..., Any]] = None,
) -> Router:
    """
    Create & return a Router that exposes:
      GET /healthz
      GET /methodz
      GET /hookz
      GET /kernelz
    """
    source_router = router
    router = Router()

    dep = get_db

    healthz_endpoint = build_healthz_endpoint(dep)

    async def _runtime_healthz(ctx: Any) -> Any:
        db = getattr(ctx, "db", None)
        if db is None:
            db = getattr(ctx, "_raw_db", None)
        if db is None:
            req = getattr(ctx, "request", None)
            state = getattr(req, "state", None)
            db = getattr(state, "db", None)
        if db is None:
            return {"ok": True}
        await maybe_execute(db, "SELECT 1")
        return {"ok": True}

    _register_runtime_paths(
        source_router,
        path="/healthz",
        alias="healthz",
        endpoint_factory=_runtime_healthz,
    )
    router.add_route(
        "/healthz",
        healthz_endpoint,
        methods=["GET"],
        name="healthz",
        tags=["system"],
        summary="Health",
        description="Database connectivity check.",
    )

    methodz_endpoint = build_methodz_endpoint(source_router)
    _register_runtime_paths(
        source_router,
        path="/methodz",
        alias="methodz",
        endpoint_factory=lambda _ctx: methodz_endpoint(),
    )
    router.add_route(
        "/methodz",
        methodz_endpoint,
        methods=["GET"],
        name="methodz",
        tags=["system"],
        summary="Methods",
        description="Ordered, canonical operation list.",
    )

    hookz_endpoint = build_hookz_endpoint(source_router)
    _register_runtime_paths(
        source_router,
        path="/hookz",
        alias="hookz",
        endpoint_factory=lambda _ctx: hookz_endpoint(),
    )
    router.add_route(
        "/hookz",
        hookz_endpoint,
        methods=["GET"],
        name="hookz",
        tags=["system"],
        summary="Hooks",
        description=(
            "Expose hook execution order for each method.\n\n"
            "Phases appear in runner order; error phases trail.\n"
            "Within each phase, hooks are listed in execution order: "
            "global (None) hooks, then method-specific hooks."
        ),
    )

    kernelz_endpoint = build_kernelz_endpoint(source_router)

    async def _runtime_kernelz(_ctx: Any) -> Any:
        from tigrbl_runtime.runtime.kernel import _default_kernel as K

        K.ensure_primed(source_router)
        payload: dict[str, dict[str, list[str]]] = {}
        for model in table_iter(source_router):
            model_name = getattr(model, "__name__", str(model))
            payload[model_name] = {}
            for sp in opspecs(model):
                payload[model_name][sp.alias] = K.plan_labels(model, sp.alias)
        return payload

    _register_runtime_paths(
        source_router,
        path="/kernelz",
        alias="kernelz",
        endpoint_factory=_runtime_kernelz,
    )
    router.add_route(
        "/kernelz",
        kernelz_endpoint,
        methods=["GET"],
        name="kernelz",
        tags=["system"],
        summary="Kernel Plan",
        description="Phase-chain plan as built by the kernel per operation.",
    )

    return router


__all__ = ["mount_diagnostics"]



from __future__ import annotations

from tigrbl_base._base import BindingBase, BindingRegistryBase


class Binding(BindingBase):
    """Concrete named binding declaration."""


class BindingRegistry(BindingRegistryBase):
    """Concrete named binding registry."""


__all__ = ["Binding", "BindingRegistry"]



from __future__ import annotations

from warnings import warn

from tigrbl_base._base._rest_map import *  # noqa: F403

warn(
    "tigrbl_concrete._mapping.rest.helpers is deprecated; use tigrbl_base._base._rest_map",
    DeprecationWarning,
    stacklevel=2,
)



# tigrbl/_concrete/_table.py
from __future__ import annotations

from types import SimpleNamespace
from typing import Any, Callable

from ._engine import AsyncSession, Session
from tigrbl_concrete.ddl import initialize as _ddl_initialize
from ._engine import Engine  # reuse the collector
from tigrbl_concrete._concrete import engine_resolver as _resolver
from tigrbl_concrete._mapping.model_helpers import _ensure_model_namespaces
from tigrbl_core._spec.table_spec import TableSpec
from tigrbl_base._base._table_base import TableBase


class Table(TableBase):
    """Declarative ORM table base.

    This class now integrates :class:`TableBase` so ORM models and tables share
    the same type. Column specifications are exposed via ``columns`` for
    convenience.
    """

    __abstract__ = True
    columns: SimpleNamespace = SimpleNamespace()

    DEFAULT_CANON_VERBS = {
        "create",
        "read",
        "update",
        "replace",
        "delete",
        "list",
        "clear",
    }

    @classmethod
    def should_wire_canonical(cls, op: str) -> bool:
        from tigrbl_core.config.constants import (
            TIGRBL_DEFAULTS_EXCLUDE_ATTR,
            TIGRBL_DEFAULTS_INCLUDE_ATTR,
            TIGRBL_DEFAULTS_MODE_ATTR,
        )

        mode = getattr(cls, TIGRBL_DEFAULTS_MODE_ATTR, "all")
        inc = set(getattr(cls, TIGRBL_DEFAULTS_INCLUDE_ATTR, set()))
        exc = set(getattr(cls, TIGRBL_DEFAULTS_EXCLUDE_ATTR, set()))
        if mode == "none":
            return False
        if mode == "some":
            return op in inc
        default_verbs = set(
            getattr(cls, "DEFAULT_CANON_VERBS", Table.DEFAULT_CANON_VERBS)
        )
        allowed = (default_verbs | inc) - exc
        return op in allowed

    @classmethod
    def _collect_mro_spec(cls) -> TableSpec:
        return TableSpec.collect(cls)

    def __init__(self, **kw: Any) -> None:  # pragma: no cover - SQLA sets attrs
        for k, v in kw.items():
            setattr(self, k, v)

    def __init_subclass__(cls, **kw: Any) -> None:  # noqa: D401
        super().__init_subclass__(**kw)

        collected_spec = cls._collect_mro_spec()

        cls.OPS = tuple(collected_spec.ops)
        cls.COLUMNS = tuple(collected_spec.columns)
        cls.SCHEMAS = tuple(collected_spec.schemas)
        cls.HOOKS = tuple(collected_spec.hooks)
        cls.SECURITY_DEPS = tuple(collected_spec.security_deps)
        cls.DEPS = tuple(collected_spec.deps)

        if collected_spec.engine is not None:
            cfg = dict(getattr(cls, "table_config", {}) or {})
            cfg.setdefault("engine", collected_spec.engine)
            cls.table_config = cfg

        # expose ColumnSpecs under `columns` namespace
        specs = getattr(cls, "__tigrbl_cols__", {})
        cls.columns = SimpleNamespace(**specs)

        # Ensure each model owns isolated namespaces for ops/schemas/hooks/etc.
        # This keeps subclass mutations from leaking across siblings.
        _ensure_model_namespaces(cls)

        # auto-register table-level bindings if declared
        try:
            Engine.install_from_objects(tables=(cls,))
        except Exception:  # pragma: no cover - best effort
            pass

    @classmethod
    def install_engines(cls, *, router: Any | None = None) -> None:
        Engine.install_from_objects(router=router, tables=(cls,))

    @classmethod
    def acquire(
        cls, *, op_alias: str | None = None
    ) -> tuple[Session | AsyncSession, Callable[[], None]]:
        db, release = _resolver.acquire(model=cls, op_alias=op_alias)
        return db, release

    initialize = classmethod(_ddl_initialize)



"""CORS middleware for ASGI/WSGI adapters."""

from __future__ import annotations

import re

from ._headers import Headers

from .base import BaseHTTPMiddleware


class CORSMiddleware(BaseHTTPMiddleware):
    _CORS_HEADER_NAMES = {
        "access-control-allow-origin",
        "access-control-allow-methods",
        "access-control-allow-headers",
        "access-control-max-age",
        "access-control-allow-credentials",
        "vary",
    }

    def __init__(
        self,
        app,
        *,
        allow_origin: str = "*",
        allow_origins: list[str] | tuple[str, ...] | None = None,
        allow_origin_regex: str | None = None,
        allow_methods: str
        | list[str]
        | tuple[str, ...] = "GET,POST,PUT,PATCH,DELETE,OPTIONS",
        allow_headers: str | list[str] | tuple[str, ...] = "*",
        allow_credentials: bool = False,
        max_age: int = 600,
    ) -> None:
        super().__init__(
            app,
            allow_origin=allow_origin,
            allow_origins=allow_origins,
            allow_origin_regex=allow_origin_regex,
            allow_methods=allow_methods,
            allow_headers=allow_headers,
            allow_credentials=allow_credentials,
            max_age=max_age,
        )
        self.allow_origin = allow_origin
        self.allow_origins = list(allow_origins or [])
        self.allow_origin_regex = allow_origin_regex
        self._allow_origin_pattern = (
            re.compile(allow_origin_regex) if allow_origin_regex else None
        )
        self.allow_methods = self._normalize_listish(allow_methods)
        self.allow_headers = self._normalize_listish(allow_headers)
        self.allow_credentials = allow_credentials
        self.max_age = max_age

    @staticmethod
    def _normalize_listish(value: str | list[str] | tuple[str, ...]) -> str:
        if isinstance(value, str):
            return value
        return ",".join(item.strip() for item in value)

    def _resolve_allow_origin(self, request_headers: Headers) -> str:
        origin = request_headers.get("origin")
        if not origin:
            return self.allow_origin

        if self._allow_origin_pattern and self._allow_origin_pattern.match(origin):
            return origin

        if self.allow_origins:
            if origin in self.allow_origins:
                return origin
            return "null"

        if self.allow_origin == "*":
            return self.allow_origin

        allowed = [candidate.strip() for candidate in self.allow_origin.split(",")]
        if origin in allowed:
            return origin
        return "null"

    def _cors_headers(self, request_headers: Headers) -> list[tuple[str, str]]:
        allow_origin = self._resolve_allow_origin(request_headers)
        allow_headers = self.allow_headers
        requested_headers = request_headers.get("access-control-request-headers")
        if allow_headers == "*" and requested_headers:
            allow_headers = requested_headers

        headers = [
            ("access-control-allow-origin", allow_origin),
            ("access-control-allow-methods", self.allow_methods),
            ("access-control-allow-headers", allow_headers),
            ("access-control-max-age", str(self.max_age)),
            ("vary", "origin"),
        ]
        if self.allow_credentials:
            headers.append(("access-control-allow-credentials", "true"))
        return headers


__all__ = ["CORSMiddleware"]



# tigrbl/v3/transport/rest/__init__.py
"""
Tigrbl v3 – REST transport wrapper.

Use this when you prefer to mount a single top-level router that aggregates all
model routers (instead of mounting each one inside include_tables).

Typical usage:
    from tigrbl.transport.rest import build_rest_router, mount_rest

    # When including models, skip mounting per-model:
    router.include_table(User, mount_router=False)
    router.include_table(Team, mount_router=False)
    # or:
    router.include_tables([User, Team], mount_router=False)

    # Then aggregate & mount once:
    app.include_router(build_rest_router(router, base_prefix="/router"))
    # or:
    mount_rest(router, app, base_prefix="/router")
"""

from __future__ import annotations

from .aggregator import build_rest_router, mount_rest

__all__ = [
    "build_rest_router",
    "mount_rest",
]



from __future__ import annotations

from typing import Any


def sqlite_capabilities(*, async_: bool, memory: bool) -> dict[str, Any]:
    return {
        "transactional": True,
        "async_native": async_,
        "isolation_levels": {"read_committed", "serializable"},
        "read_only_enforced": False,
        "supports_timeouts": True,
        "supports_ddl": True,
        "engine": "sqlite",
        "memory": bool(memory),
    }


def postgres_capabilities(*, async_: bool) -> dict[str, Any]:
    return {
        "transactional": True,
        "async_native": async_,
        "isolation_levels": {"read_committed", "repeatable_read", "serializable"},
        "read_only_enforced": True,
        "supports_timeouts": True,
        "supports_ddl": True,
        "engine": "postgres",
    }



"""Utilities for running uvicorn during tests or tooling."""

from __future__ import annotations

import asyncio

import uvicorn


async def _cancel_task(task: asyncio.Task) -> None:
    if task.done():
        return
    task.cancel()
    try:
        await task
    except asyncio.CancelledError:
        return


async def _close_servers(server: uvicorn.Server) -> None:
    servers = []
    primary = getattr(server, "server", None)
    if primary is not None:
        servers.append(primary)
    extra = getattr(server, "servers", None)
    if extra:
        servers.extend(extra)
    for srv in servers:
        close = getattr(srv, "close", None)
        if callable(close):
            close()
        wait_closed = getattr(srv, "wait_closed", None)
        if callable(wait_closed):
            await wait_closed()


async def stop_uvicorn_server(
    server: uvicorn.Server,
    task: asyncio.Task,
    *,
    timeout: float = 5.0,
) -> None:
    """Request uvicorn shutdown and ensure the task exits."""
    if task.done():
        return

    server.should_exit = True
    try:
        await asyncio.wait_for(task, timeout=timeout)
        return
    except asyncio.TimeoutError:
        server.force_exit = True
        shutdown = getattr(server, "shutdown", None)
        if callable(shutdown):
            try:
                await asyncio.wait_for(shutdown(), timeout=timeout)
            except asyncio.TimeoutError:
                pass
        await _close_servers(server)
        await _cancel_task(task)
